import random
import os
import wandb
import json
import numpy as np
import torch
import logging
from torch.utils.data import DataLoader, IterableDataset, get_worker_info
from typing import List, Tuple, Dict, Any, Literal
import torchvision.transforms.functional as tvF
import tiffslide as ts
from pathlib import Path
import itertools
from tqdm.notebook import tqdm
from sklearn.metrics import accuracy_score, f1_score


from mmm.utils import convert_tile_into_patchbatch, flatten_list_of_dicts
from mmm.mtl_modules.tasks.ClassificationTask import ClassificationTask
from mmm.mtl_modules.shared_blocks.PyramidEncoder import PyramidEncoder
from mmm.interactive import blocks, configs as cfs, data, pipes, tasks, training
from mmm.mtl_modules.shared_blocks.SharedModules import SharedModules


class SemiColConfig(cfs.BaseModel):
    write: bool = False
    alternative_path: str = "/output/tissueconcepts/nics/only-nics"
    level: int = 0
    patch_size: int = 224


class TupacConfig(cfs.BaseModel):
    write: bool = False
    alternative_path: str = "/output/tissueconcepts/nics/only-nics"
    level: int = 1
    patch_size: int = 224


class NICConfig(cfs.BaseModel):
    semicol_config: SemiColConfig = SemiColConfig()
    tupac_config: TupacConfig = TupacConfig()
    head_config: blocks.PyramidEncoder.Config = blocks.PyramidEncoder.Config(
        model=cfs.MiniConvNetConfig(num_channels=768),
        # Because of varying input sizes the batchsize is 1 and the accumulation is large.
        # High gradient accumulation means batchnorm is a bad choice
        norm_layer="affinelayernorm",
        # activation_fn=cfs.ActivationFunctionConfig(fn_type=cfs.ActivationFn.ReLU),
        # hidden_dim=64,
        module_name="encoder",
    )


class NICTask(ClassificationTask):
    """
    A classification task that processes neural image representations [C, H, W].
    """

    class Config(ClassificationTask.Config):
        pass

    def _visualize_preds(self, training_ims, step_metrics: Dict, metas: List[Dict]) -> Dict[str, Any]:
        vis_n = min(self._takeout_vis_budget(), training_ims.size(0))

        if vis_n <= 0:
            return {}

        nic = training_ims[0]
        meta = metas[0]
        metastr = json.dumps(meta, default=lambda o: str(o))

        logging.info(nic.shape)
        fmap_index = random.randint(0, nic.shape[0] - 1)

        logging.info(metas)

        return {"fmap": wandb.Image(nic[fmap_index], caption=metastr)}

    def training_step(self, batch: Dict[str, Any], shared_blocks: SharedModules):
        x = batch["image"]
        y = batch["class"]
        # skip if batch is empty
        tensornums = [bool(t.numel()) for t in x]
        if not True in tensornums:
            logging.info("encountered batch without valid training examples, skipping batch")
            return None

        y_hat = shared_blocks.forward(x, self.forward)

        # dealing last batch of a Cachingsubcase Dataset
        assert y.shape[0] == y_hat.shape[0], f"Till found something, {y.shape=}, {y_hat.shape=}"
        # if not y.shape[0] == y_hat.shape[0]:
        #     y = y.long()
        #     y_hat = y_hat.unsqueeze(0)

        batch_loss = self.criterion(y_hat, y) / np.log(len(self.class_names))
        split = np.asarray([b["split"] for b in batch["meta"]])
        step_results: StepMetricDict = {  # type: ignore (.numpy() does not correctly indicate numpy array)
            "targets": y.cpu().numpy(),
            "logits": y_hat.detach().cpu().numpy(),
            "preds": torch.argmax(y_hat.detach().cpu(), dim=1).numpy(),
            "lab": split,
        }
        self.add_step_result(batch_loss.item(), step_results)

        live_vis = self._visualize_preds(
            x.detach().cpu(),
            step_results,
            (batch["meta"] if "meta" in batch else [{} for _ in range(batch["image"].shape[0])]),
        )

        return batch_loss, live_vis

    def log_epoch_metrics(self) -> Tuple[Dict[str, Any], str]:
        log_dict, print_str = super().log_epoch_metrics()
        metrics = flatten_list_of_dicts(self._step_metrics)

        # Change such that we log the metrics overall and per split
        labs = np.unique(metrics["lab"])
        for lab in labs:
            log_dict[f"{lab}_acc"] = accuracy_score(
                y_true=metrics["targets"][np.where(metrics["lab"] == lab)],
                y_pred=metrics["preds"][np.where(metrics["lab"] == lab)],
            )
            log_dict[f"{lab}_f1"] = f1_score(
                y_true=metrics["targets"][np.where(metrics["lab"] == lab)],
                y_pred=metrics["preds"][np.where(metrics["lab"] == lab)],
                average="macro",
            )
            log_dict[f"{lab}_confmat"] = wandb.plot.confusion_matrix(
                preds=metrics["preds"][np.where(metrics["lab"] == lab)],  # type: ignore
                y_true=metrics["targets"][np.where(metrics["lab"] == lab)],  # type: ignore
                class_names=self._get_short_class_names(),
                title=f"{self._prefix}_{self.get_name()}_{lab}",
            )

        return log_dict, print_str


def get_next_divisable(n: int, divisor: int, updown: Literal["larger", "smaller"] = "larger") -> int:
    if updown == "larger":
        return n + (divisor - n % divisor) % divisor
    elif updown == "smaller":
        return n - n % divisor


def get_tile_edge_length(pixels_per_chunk: int = 96 * 224 * 224, patch_edge_length: int = 224) -> int:
    max_edge_length = np.sqrt(pixels_per_chunk)
    divisable_edge_length = int(max_edge_length - (max_edge_length % patch_edge_length))
    return divisable_edge_length


def compute_positions(dimsizes, stepsize):
    return itertools.product(*[enumerate(range(0, dimsize, stepsize)) for dimsize in dimsizes])


def read_regions(
    slide: ts.TiffSlide,
    positions: list,
    tile_edgelength: int,
    patch_edge_length: int,
    level=0,
):
    for (row, x), (col, y) in positions:
        dim1_edgelength = min(
            tile_edgelength,
            get_next_divisable(slide.level_dimensions[level][0] - x, patch_edge_length, "larger"),
        )
        dim2_edgelength = min(
            tile_edgelength,
            get_next_divisable(slide.level_dimensions[level][1] - y, patch_edge_length, "larger"),
        )
        location_downsample_fac = slide.level_downsamples[level]

        logging.debug(
            f"Region {row=}, {col=}, {x=}, {y=}, {dim1_edgelength=}, {dim2_edgelength=}, {level=}, {location_downsample_fac=}"
        )
        region = tvF.to_tensor(
            slide.read_region(
                location=(
                    int(x * location_downsample_fac),
                    int(y * location_downsample_fac),
                ),
                level=level,
                size=(dim1_edgelength, dim2_edgelength),
                as_array=True,  # seems to be a little faster than converting from PIL
            )
        )
        # region = build_test_tensor(dim1_edgelength, dim2_edgelength)
        assert region.shape[0] == 3, "Expected RGB image, after to_tensor it should be channels first"
        yield (row, x), (col, y), region


def convert_region_to_repr(
    enc: PyramidEncoder,
    region: torch.Tensor,
    patch_size: int,
    stride: int,
    from_repr: Literal["pyramid", "z"] = "z",
):
    batch, rows, cols = convert_tile_into_patchbatch(region, patch_size=patch_size, stride=stride)

    with torch.inference_mode():
        pyr, z = enc(batch.to(next(enc.parameters()).device))
        # Here you can encode debug information
        # z[:, 0] = x
        # z[:, 1] = y

    repr = pyr[-1] if from_repr == "pyramid" else z

    # Convert to rows, cols, channels, ...
    repr = repr.reshape(rows, cols, *repr.shape[1:]).swapaxes(1, 0)
    return repr


class RegionLoader(IterableDataset):
    def __init__(
        self,
        slide_paths: List[Path],
        extraction_level: int,
        patch_size: int,
        max_pixels_per_chunk: int,
    ):
        self.slide_paths, self.level, self.patch_size = (
            slide_paths,
            extraction_level,
            patch_size,
        )
        self.max_pixels_per_chunk = max_pixels_per_chunk

    def __iter__(self):
        worker_info = get_worker_info()

        # First move: find out which supercases this worker should process
        if worker_info is None:
            my_slide_paths = self.slide_paths
            worker_id = 0
        else:
            num_workers = worker_info.num_workers
            worker_id: int = worker_info.id
            my_slide_paths = self.slide_paths[worker_id::num_workers]
            logging.debug(
                f"Worker {worker_id} got {len(my_slide_paths)} supercases: {[x.stem for x in my_slide_paths]}"
            )

        for slide_path in my_slide_paths:
            slide = ts.TiffSlide(slide_path)
            # tf = tifffile.TiffFile(slide_path)

            num_regions_of_this_slide = 0
            tile_edgelength = get_tile_edge_length(
                patch_edge_length=self.patch_size,
                pixels_per_chunk=self.max_pixels_per_chunk,
            )
            positions = list(compute_positions(slide.level_dimensions[self.level], tile_edgelength))
            if worker_id == 0:
                logging.debug(f"I need to process {len(positions)} windows")
            try:
                for x in read_regions(slide, positions, tile_edgelength, self.patch_size, level=self.level):
                    yield slide_path, x
                    num_regions_of_this_slide += 1
            except Exception as e:
                logging.warning(f"Error while reading slide {slide_path}: {e}")

            if worker_info is None or worker_id == 0:
                logging.debug(f"There were {num_regions_of_this_slide} regions in slide {slide_path}")
                num_regions_of_this_slide = 0


def load_representations_for_tiffs(
    enc: PyramidEncoder,
    tiffpaths: List[Path],
    num_workers=4,
    level=0,
    patch_size=224,
    max_pixels_per_batch=96 * 224 * 224,
) -> Dict:
    region_ds = RegionLoader(tiffpaths, level, patch_size, max_pixels_per_batch)
    loader = DataLoader(region_ds, collate_fn=lambda x: x[0], batch_size=1, num_workers=num_workers)
    mem = {}
    for slide_path, ((row, x), (col, y), region) in tqdm(loader):
        if slide_path not in mem:
            mem[slide_path] = {}
        mem[slide_path][(row, col)] = (
            x,
            y,
            convert_region_to_repr(enc, region, patch_size, patch_size),
        )

    return mem


def build_nic(tiff_repr: Dict[Tuple[int, int], Tuple[int, int, torch.Tensor]]) -> torch.Tensor:
    """
    Based on the input dictionary which maps a tile position to a representation tensor,
    build the neural image representation of the whole slide.

    The representation can be converted to channels first by calling `nic.permute(2, 0, 1)`

    Confirm using:

    ```
    x = torch.Tensor([[[0.,  1.,  2.],
                    [3.,  4.,  5.]],

                    [[6.,  7.,  8.],
                    [9., 10., 11.]]])
    assert torch.equal(x[0, 1], x.permute(2, 0, 1)[:, 0, 1]) and torch.equal(x.permute(2, 0, 1)[:, 0, 1], torch.Tensor([3., 4., 5.]))
    ```
    """
    num_cols = [repr.shape[1] for (row, col), (x, y, repr) in tiff_repr.items() if row == 0]
    num_rows = [repr.shape[0] for (row, col), (x, y, repr) in tiff_repr.items() if col == 0]
    depth = tiff_repr[(0, 0)][2].shape[2]

    nic = torch.zeros(size=(sum(num_rows), sum(num_cols), depth), dtype=tiff_repr[(0, 0)][2].dtype)

    for (row, col), (x, y, repr) in tiff_repr.items():
        r = row * num_rows[0]
        c = col * num_cols[0]
        nic[r : r + repr.shape[0], c : c + repr.shape[1], :] = repr

    return nic


def process_tile(nic: torch.Tensor):
    """
    Removes the last row and column of the image and corrects the orientation.

    Probably SemiCOL challenge specific.
    """
    mirrored = nic.rot90(3).flip(1).permute(2, 0, 1)[:, :-1, :-1]

    return mirrored


def get_output_path_of_tile(in_path: Path, outfolder: Path) -> Path:
    return outfolder / in_path.parent.parts[-1] / f"{in_path.name}.pt"


def write_nic_from_list(
    trainfiles: List[Path],
    outfolder: Path,
    encoder: blocks.PyramidEncoder,
    level: int,
    patch_size: int,
    num_workers: int,
):
    """
    Function to write NICs from a list of absolute paths to a files. The files should be supported by the tiffslide package
    https://pypi.org/project/tiffslide/

    For now, we are operating with .tif and .svs files, which both are fully supported by the library.
    """

    while not_yet_existing := [
        filepath for filepath in trainfiles if not get_output_path_of_tile(filepath, outfolder).exists()
    ]:
        filepaths = random.sample(not_yet_existing, min(len(not_yet_existing), num_workers))
        representations = load_representations_for_tiffs(
            encoder,
            filepaths,
            num_workers=num_workers,
            patch_size=patch_size,
            level=level,
            max_pixels_per_batch=96 * 3 * patch_size * patch_size,
        )
        nics = {p: build_nic(representations[p]) for p in representations.keys()}

        for p, nic in nics.items():
            out_path = get_output_path_of_tile(p, outfolder)
            if not out_path.exists():
                out_path.parent.mkdir(parents=True, exist_ok=True)
                torch.save(nic, out_path)



from mmm.transforms import (
    RandomApply,
    KeepOnlyKeysInDict,
    ExtractKeysAsTuple,
    TupleToDict,
    CopyKeysInDict,
    UnifySizes,
    ApplyToKey,
    ApplyToKeys,
    ApplyToList,
    SaveResultAsKey,
    ChannelsSwapFirstLast,
    ResizeWithMask,
    ResizeImage,
    ResizeBoxes,
    MaskedPatchExtractor,
    batchify,
    Alb,
    AlbWithBoxes,
)
from mmm.augmentations import (
    get_histo_augs,
    get_xray_augs,
    get_weak_default_augs,
    get_realworld_augs,
    get_color_prediction_transform,
    get_surrounding_prediction_transform,
)
from mmm.data_loading.MTLDataset import mtl_collate, mtl_batch_collate



import torch
import torch.nn as nn
import torch.nn.functional as F


class StateLessBatchnorm(nn.BatchNorm2d):
    """
    Adaption of original batchnorm which should work without tracking stats.
    Architecture that use it need to provide full batches for inference.
    """

    def __init__(self, num_features, eps=0.00001, momentum=0.1, affine=True, device=None):
        track_running_stats = False
        dtype = None
        super().__init__(num_features, eps, momentum, affine, track_running_stats, device, dtype)

    @staticmethod
    def from_batchnorm(batchnorm: nn.BatchNorm2d):
        return StateLessBatchnorm(batchnorm.num_features, batchnorm.eps, batchnorm.momentum, batchnorm.affine)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        # self._check_input_dim(input)

        # if self.training and self.track_running_stats:
        #     # TODO: if statement only here to tell the jit to skip emitting this when it is None
        #     if self.num_batches_tracked is not None:  # type: ignore[has-type]
        #         self.num_batches_tracked.add_(1)  # type: ignore[has-type]
        #         if self.momentum is None:  # use cumulative moving average
        #             exponential_average_factor = 1.0 / float(self.num_batches_tracked)
        #         else:  # use exponential moving average
        #             exponential_average_factor = self.momentum
        r"""
        Decide whether the mini-batch stats should be used for normalization rather than the buffers.
        Mini-batch stats are used in training mode, and in eval mode when buffers are None.
        """
        if self.training:
            bn_training = True
        else:
            bn_training = (self.running_mean is None) and (self.running_var is None)

        r"""
        Buffers are only updated if they are to be tracked and we are in training mode. Thus they only need to be
        passed when the update should occur (i.e. in training mode when they are tracked), or when buffer stats are
        used for normalization (i.e. in eval mode when buffers are not None).
        """
        return F.batch_norm(
            input,
            None,  # No tracked mean
            None,  # No Tracked variance
            self.weight,
            self.bias,
            bn_training,
            self.momentum,
            self.eps,
        )



from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.mtl_modules.shared_blocks.SharedModules import SharedModules
from mmm.mtl_modules.shared_blocks.PyramidEncoder import PyramidEncoder
from mmm.mtl_modules.shared_blocks.MTLDecoder import MTLDecoder
from mmm.mtl_modules.shared_blocks.Squeezer import Squeezer
from mmm.mtl_modules.shared_blocks.Grouper import Grouper

from mmm.neural.modules.simple_cnn import MiniConvNet, MiniDecoder
from mmm.neural.modules.swinformer import TorchVisionSwinformer
from mmm.neural.modules.TorchVisionCNN import TorchVisionCNN

from mmm.neural.modules.smp_modules import SMPUnetDecoder
from mmm.mtl_modules.shared_blocks.PyramidDecoder import PyramidDecoder

# Optional dependency group for detection
try:
    from mmm.mtl_modules.shared_blocks.FCOSDecoder import FCOSDecoder
except ImportError:
    FCOSDecoder = None

try:
    from mmm.mtl_modules.shared_blocks.TextEmbedder import TextEmbedder
except ImportError:
    TextEmbedder = None



import logging
import re
from typing import Callable
from pydantic import Field
from mmm.BaseModel import BaseModel

from .GeoAnno import GeoAnno


class AnnoFilterConfig(BaseModel):
    """
    Filters annotations by regex. All annotations without a class name are filtered by default.
    """

    delete_regex: str | None = Field(
        default=None,
        description="Only include annotations with class names that do not match",
    )
    keeponly_regex: str | None = Field(
        default=None,  # f".*({'|'.join([x[0] for x in recommended_buckets])}).*",
        description="Only include annotations with class names that match",
    )

    def apply(
        self,
        x: GeoAnno,
        additional_function: Callable[[GeoAnno], bool] | None = None,
        use_regions: list[GeoAnno] | None = None,
        ignore_regions: list[GeoAnno] | None = None,
    ) -> bool:
        """
        wsi_regions filters annotations that do not intersect with any of the wsi_regions
        """
        anno_class_name = x.get_class_name()

        if anno_class_name is None:
            logging.debug(f"omitting {x.shape.bounds}, {x.get_class_name()} because no class")
            return False

        anno_class_name = anno_class_name.lower()

        if self.delete_regex is not None:
            if re.match(self.delete_regex, anno_class_name):
                logging.debug(f"omitting {x.shape.bounds}, {x.get_class_name()} because {self.delete_regex=}")
                return False

        if self.keeponly_regex is not None:
            if not re.match(self.keeponly_regex, anno_class_name):
                logging.debug(f"omitting {x.shape.bounds}, {x.get_class_name()} because {self.keeponly_regex=}")
                return False

        if use_regions is not None:
            if not any([x.shape.intersects(y.shape) for y in use_regions]):
                logging.debug(f"omitting {x.shape.bounds}, {x.get_class_name()} because not in wsi regions")
                return False

        if ignore_regions is not None:
            if any([x.shape.intersects(y.shape) for y in ignore_regions]):
                logging.debug(f"omitting {x.shape.bounds}, {x.get_class_name()} because in ignore regions")
                return False

        # Should be the last statement
        return additional_function is None or additional_function(x)



import torch

from .SharedBlock import SharedBlock

try:
    from transformers import AutoTokenizer, CLIPTextModel
except ImportError:
    AutoTokenizer, CLIPTextModel = None, None


class TextEmbedder(SharedBlock):
    class Config(SharedBlock.Config):
        module_name: str = "textembedder"
        modelname: str = "openai/clip-vit-base-patch32"
        # pretrained: str = "laion2b_s34b_b79k"

    def __init__(self, cfg: Config):
        super().__init__(cfg)
        self.model = CLIPTextModel.from_pretrained(cfg.modelname)
        self._tokenizer = AutoTokenizer.from_pretrained(cfg.modelname)
        self.make_mtl_compatible()

    def get_hidden_dim(self):
        return self.model.text_model.final_layer_norm.normalized_shape[0]

    def tokenize(self, texts: list[str]) -> dict[str, torch.Tensor]:
        """returns a dict of tensors with keys: input_ids, attention_mask"""
        return self._tokenizer(texts, padding=True, return_tensors="pt")

    def forward(self, input_tokens: torch.Tensor, attention_mask: torch.Tensor):
        outputs = self.model(input_ids=input_tokens, attention_mask=attention_mask)
        # outputs.last_hidden_state exists as well
        return outputs.pooler_output



import torch
import json
import numpy as np
from typing import Any
from pathlib import Path
from torch.utils.data import Dataset
import torchvision.transforms.functional as F
import imageio.v3 as imageio
from torchvision.datasets.folder import default_loader


class DiskDataset(Dataset):
    def __init__(self, folder_path, meta_file="mtlfiledataset.json"):
        self.folder_path = folder_path
        self.case_paths: list[Path] = self.update()
        self.dataset_meta: dict = json.loads((self.folder_path / meta_file).read_text())

    def update(self) -> list[Path]:
        self.case_paths = list(filter(lambda x: x.is_dir(), self.folder_path.iterdir()))
        return self.case_paths

    def __len__(self):
        self.update()
        return len(self.case_paths)

    def __getitem__(self, index) -> dict[str, Any]:
        folder_path = self.case_paths[index]
        res = {}
        # Load jpg as Tensors
        for img_path in folder_path.glob("*.jpg"):
            res[img_path.stem] = F.to_tensor(default_loader(img_path))
        # Load npz using numpy
        for npz_path in folder_path.glob("*.npz"):
            res[npz_path.stem] = torch.from_numpy(np.load(npz_path)["arr_0"])
        # Load case.json into meta key
        res["meta"] = json.load((folder_path / "case.json").open())
        res["meta"]["case_path"] = str(folder_path)
        return res



from __future__ import annotations
import wandb
import re
import logging
import json
import json5
from typing import List
from pathlib import Path
import os
from typing import Optional

from pydantic import Field, model_validator
from mmm.BaseModel import BaseModel

# Dataloading
from mmm.data_loading.TrainValCohort import TrainValCohort
from mmm.logging.wandb_ext import remove_wandb_special_chars
from mmm.DataSplit import DataSplit

# Training
from mmm.trainer.Loop import LoopConfig, TrainLoopConfig, ValLoopConfig
from mmm.trainer.MTLTrainer import MTLTrainer, EarlyStoppingConfig
from mmm.optimization.MTLOptimizer import (
    MTLOptimizer,
    OptimizerAdamWConfig,
    OptimizerSGDConfig,
    SchedulerType,
)

from mmm.optimization.MTLOptimizer import (
    PolySchedulerConfig,
    ExponentialLRConfig,
    CosineAnnealingLRSchedulerConfig,
    DecaySchedulerConfig,
    ReduceLROnPlateauConfig,
)
from mmm.event_selectors import (
    FixedEventSelector,
    CombinedEventSelector,
    RecurringEventSelector,
)

# Blocks
from mmm.mtl_modules.shared_blocks.PyramidEncoder import PyramidEncoder
from mmm.mtl_modules.shared_blocks.PyramidDecoder import PyramidDecoder
from mmm.mtl_modules.shared_blocks.FCOSDecoder import FCOSDecoder
from mmm.mtl_modules.shared_blocks.MTLDecoder import MTLDecoder
from mmm.neural.modules.smp_modules import SMPPyramidAttentionDecoderConfig
from mmm.neural.modules.TorchVisionCNN import TorchVisionCNN

from mmm.neural.modules.swinformer import TorchVisionSwinformer
from mmm.neural.modules.simple_cnn import MiniConvNet
from mmm.neural.modules.convnext import TorchVisionConvnext
from mmm.neural.modules.TimmEncoder import TimmEncoder
from mmm.neural.activations import ActivationFunctionConfig, ActivationFn
from mmm.neural.pooling import GlobalPoolingConfig, GlobalPooling

# Tasks
from mmm.mtl_modules.tasks.MTLTask import MTLTask
from mmm.mtl_modules.tasks.ClassificationTask import ClassificationTask
from mmm.mtl_modules.tasks.RegressionTask import RegressionTask
from mmm.mtl_modules.tasks.SemSegTask import SemSegTask
from mmm.mtl_modules.tasks.mmDetectionTask import MMDetectionTask
from mmm.mtl_modules.tasks.MultilabelClassificationTask import (
    MultilabelClassificationTask,
)

TrainValCohortConfig = TrainValCohort.Config
# Training
MTLOptimizerConfig = MTLOptimizer.Config

# Neural
from mmm.neural.pooling import GlobalPoolingConfig, GlobalPooling

# Blocks
MiniConvNetConfig = MiniConvNet.Config
TorchVisionSwinformerConfig = TorchVisionSwinformer.Config
FCOSDecoderConfig = FCOSDecoder.Config
AEDecoderConfig = MTLDecoder.Config
PyramidEncoderConfig = PyramidEncoder.Config
PyramidDecoderConfig = PyramidDecoder.Config
TorchVisionConvnextConfig = TorchVisionConvnext.Config

# Tasks
MTLTaskConfig = MTLTask.Config
ClassificationTaskConfig = ClassificationTask.Config
SemSegTaskConfig = SemSegTask.Config
MMDetectionTaskConfig = MMDetectionTask.Config
MultilabelClassificationTaskConfig = MultilabelClassificationTask.Config


class EnvByConvention:
    """
    Loads info about your environment by the conventions.
    """

    def __init__(self, env_name: str = "training") -> None:
        """
        `env_name` might be "notebookname" for a notebook with filename `notebookname.ipynb`
        """
        from pathlib import Path

        # This is needed to override the logging of IPython
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)

        self.env_name = env_name
        self.data_root = Path(os.getenv("ML_DATA_ROOT", default="/data_root/"))
        self.data_cache = Path(os.getenv("ML_DATA_CACHE", default="/dl_cache/"))
        self.data_output = Path(os.getenv("ML_DATA_OUTPUT", default="/data_output/"))
        self.juicefs_mount = Path(os.getenv("ML_JUICEFS_MOUNT", default="/jfs/"))
        self.interactive_environment = os.getenv("LOCAL_DEV_ENV", default="False") == "True"
        self.job_config_folder: Path = Path("./job_configs/")
        self.rank = int(os.getenv("RANK", default="0"))
        self.world_size = int(os.getenv("WORLD_SIZE", default="1"))

    def get_default_jobconfig_path(self) -> Path:
        return self.job_config_folder / f"{self.env_name}.jsonc"

    def get_schema_path(self) -> Path:
        return Path(f"./.vscode/{self.env_name}.schema.json")

    def if_torchrun_prepare(self) -> EnvByConvention:
        """
        Determines if the torchrun environment is set up. If so, will prepare dist training if applicable.
        """
        from mmm.utils import find_missing_torchrun_envvars
        import torch
        import torch.distributed as dist

        disttraining = len(find_missing_torchrun_envvars()) == 0
        if not disttraining:
            logging.info(f"Assuming local training because {find_missing_torchrun_envvars()} are not set")
        else:
            dist.init_process_group(backend="nccl")
        torch.cuda.set_device(int(os.getenv("LOCAL_RANK", default=0)))
        return self

    def __repr__(self) -> str:
        res = f"{self.data_root=}\n{self.data_output=}\n{self.juicefs_mount=}\n"
        res += f"{self.data_cache=}\n{self.interactive_environment=}\n"
        return res


class ExperimentHyperParameters(BaseModel):
    """ """

    experiment_name: str = Field(
        default="default",
        help="Multiple runs can have the same name. \
        Group name can be explicitly set by dot notation. \
        Example: 'name', 'group.name'. You can usually correct the name in W&B without consequences. \
        ",
    )
    resumable: bool = Field(
        default_factory=lambda: os.getenv("LOCAL_DEV_ENV", default="False") == "False",
        help="If true, \
              this job will automatically resume if it is stopped or crashes.\
              It will infer the ID from the experiment's name, you will have to find it from the W&B GUI.",
    )

    # Optional
    wandb_project: str | None = None
    wandb_entity: str | None = None
    wandb_notes: str | None = Field(default="", description="Make some notes about your experiment here")
    wandb_host: str | None = Field(
        default=None,
        description="Useful to identify the physical machine that your job runs on",
    )
    wandb_save_code: bool = Field(
        default=False,
        description="If False, WANDB will still save the commit. If true, will save the entry file's code.",
    )
    wandb_job_type: str | None = None

    clear_cache_on_node: str | None = Field(
        default=None,
        description="If this matches the node name (regex), env.data_cache will be cleared.",
    )

    # Top-level configs starting with "example_" do not get sent to W&B
    example_cnn_optim: MTLOptimizer.Config = Field(
        default=MTLOptimizerConfig(optim_config=OptimizerAdamWConfig(lr=2.5e-4)),
        description="Optimizer settings that work well with CNNs such as Resnet and Densenet.",
    )
    example_vit_optim: MTLOptimizer.Config = Field(
        default=MTLOptimizerConfig(optim_config=OptimizerAdamWConfig(lr=1e-4)),
        description="Optimizer settings that work well with vision transformers such as swin transformer.",
    )
    example_schedulers_cosinedecay: List[SchedulerType] = [
        DecaySchedulerConfig(last_epoch=100),
        CosineAnnealingLRSchedulerConfig(last_epoch=100),
    ]

    example_encoder_densenet_imgnet: PyramidEncoder.Config = PyramidEncoder.Config(
        model=TorchVisionCNN.Config(pretrained=True)
    )  # type: ignore

    example_encoder_swintiny_imgnet: PyramidEncoder.Config = PyramidEncoder.Config(
        model=TorchVisionSwinformer.Config(pretrained=True),
    )  # type: ignore

    example_encoder_convnext_imgnet: PyramidEncoder.Config = PyramidEncoder.Config(
        model=TimmEncoder.Config(
            pretrained=True,
            variant="convnext_tiny",
        ),
    )

    @staticmethod
    def load_config_string(env: EnvByConvention) -> str:
        import json

        if not env.interactive_environment:
            config_env_name = f"MLOPS_JSON_{env.env_name}"
            assert config_env_name in list(
                os.environ.keys()
            ), f"""
For non-interactive jobs, the job config must be provided using the environment variable {config_env_name}.
            """
            return os.getenv(config_env_name, default="{}")
        else:
            config_path = env.get_default_jobconfig_path()
            if not config_path.exists():
                config_path.parent.mkdir(exist_ok=True)
                config_path.write_text(r"{}")
                logging.info(
                    f"""
    For {env.interactive_environment=} jobs, the job config must be provided in {env.job_config_folder}.
    You can override the default location using the {type(env).__name__} constructor.
    The file was automatically created for you at {config_path}.
                """
                )
            with open(config_path, "r") as f:
                return json.dumps(json5.load(f))

    @classmethod
    def load_config(cls, env: EnvByConvention):
        from mmm.utils import load_config_from_str

        config = load_config_from_str(cls, ExperimentHyperParameters.load_config_string(env))
        return config

    @classmethod
    def update_schema(cls, env: EnvByConvention):
        if not env.get_schema_path().exists():
            config_file_path = env.get_default_jobconfig_path()
            logging.info(
                f"""
Creating schema file at {env.get_schema_path()}.
You can use this to get autocompletion in VSCode by appending to json.schemas like
{{
    "fileMatch": ["{config_file_path}"],
    "url": "{env.get_schema_path()}"
}}
"""
            )
        env.get_schema_path().parent.mkdir(exist_ok=True, parents=True)
        env.get_schema_path().write_text(json.dumps(cls.model_json_schema()))

    @staticmethod
    def _delete_dir(direc: Path):
        for p in direc.iterdir():
            if p.is_dir():
                logging.debug(f"Removing directory {p}")
                ExperimentHyperParameters._delete_dir(p)
            else:
                p.unlink()
        direc.rmdir()

    def clear_cache_on_nodes_like(self, env: EnvByConvention, machine_id: Optional[str]):
        if self.clear_cache_on_node is not None and machine_id is not None:
            if not env.data_cache.exists():
                logging.warn(f"not delete cache, because {env.data_cache} does not exist")
                return
            if re.match(self.clear_cache_on_node, machine_id):
                import shutil

                assert "mtltorch" in env.data_cache.parts
                free_before = shutil.disk_usage(env.data_cache)[2]
                logging.info(f"Clearing cache at {env.data_cache}")
                for p in env.data_cache.iterdir():
                    if p.is_dir():
                        print(f"Removing directory {p}")
                        self._delete_dir(p)
                    else:
                        p.unlink()
                free_after = shutil.disk_usage(env.data_cache)[2]
                logging.info(f"Freed {(free_after-free_before)/1024/1024/1024:.2f} GB")

            else:
                logging.warn(
                    f"Cache will only be cleared on nodes matching {self.clear_cache_on_node}, this is {machine_id}"
                )

    def get_unique_id(self, rank: int, world_size: int, experiment_name: str) -> str:
        """
        rank refers to the global rank.
        """
        if world_size == 1:
            uid = f"{remove_wandb_special_chars(experiment_name)}"
        else:
            uid = f"{remove_wandb_special_chars(experiment_name)}{rank}"

        if uid == "default":
            uid = f"{uid}-{wandb.util.generate_id()}"
        return uid

    def get_run_description(self, rank: int, world_size: int) -> str | None:
        if "." in self.experiment_name:
            # n = self.get_unique_id(rank, world_size).split(".")[1]
            assert len(self.experiment_name.split(".")) == 2, f"{self.experiment_name=} must have max one dot"
            n = self.get_unique_id(rank, world_size, self.experiment_name.split(".")[1])
        else:
            n = self.get_unique_id(rank, world_size, self.experiment_name)

        return f"{n[:None if world_size == 1 else -1]} {rank+1}/{world_size}"

    def get_groupname(self, rank: int, world_size: int):
        if "." in self.experiment_name:
            return remove_wandb_special_chars(self.experiment_name.split(".")[0])
        if self.resumable:
            return remove_wandb_special_chars(self.experiment_name)
        else:
            return "temporary"

    def init_experiment(self, env: EnvByConvention, always_restart: bool = False):
        if wandb.run is None or always_restart:
            if machine_id := os.getenv("NODE_NAME", default=self.wandb_host):
                os.environ["WANDB_HOST"] = machine_id
            wandb_run = wandb.init(
                project=self.wandb_project,
                entity=self.wandb_entity,
                id=self.get_unique_id(env.rank, env.world_size, self.experiment_name),
                group=self.get_groupname(env.rank, env.world_size),
                job_type=self.wandb_job_type,
                name=self.get_run_description(env.rank, env.world_size),
                save_code=self.wandb_save_code,
                config={k: v for k, v in self.model_dump().items() if not k.startswith("example_")},
                # If started not resumable and then it overwrites the data of the runs with the same ID
                resume=None if not self.resumable else "allow",
            )
            wandb.log(
                {
                    "report/config": wandb.Html(
                        f"""
                <h3>The custom config for this training:</h3>
                <pre><code>{json.dumps(json.loads(self.load_config_string(env)), indent=2)}</pre></code>
                """
                    )
                }
            )
            logging.info(f"Run resumed? {wandb_run.resumed=}")
            if wandb.run.resumed and not self.resumable:
                raise Exception("Resume not allowed by config")

            self.clear_cache_on_nodes_like(env, machine_id)
        else:
            logging.info(f"wandb already initialized, not starting a new run because {always_restart=}")
            wandb_run = wandb.run
        return wandb_run



import logging
from pydantic import Field
from typing import TYPE_CHECKING, Any
from deepdiff import DeepDiff


from mmm.BaseModel import BaseModel

try:
    from label_studio_sdk import Client, Project, parse_config
except ImportError:
    if not TYPE_CHECKING:
        Client, Project = Any, Any


class LSProject:
    """
    Creating data happens by using LSProject.ls.import_tasks.
    """

    class Config(BaseModel):
        name: str = Field(
            default="uniqueprojectname",
            description="Unique identifier of the project w.r.t. this labelstudio instance",
        )

    def __init__(
        self,
        args: Config,
        ls_client: Client,
        template: str = "",
        ls_project: Project | None = None,
    ):
        self.args, self.client = args, ls_client
        self.template = template
        # If project already exists, load it
        self.ls: Project = self.get_project() if ls_project is None else ls_project
        if self.ls is None:
            self.ls = self._create_new()

    @staticmethod
    def exists(client, project_name: str) -> bool:
        return any(p.title == project_name for p in client.get_projects())

    def get_project(self) -> Project | None:
        for p in self.client.get_projects():
            if p.title == self.args.name:
                if not self.template:
                    self.template = p.get_params()["label_config"]
                else:
                    if differences := DeepDiff(
                        parse_config(self.template),
                        p.get_params()["parsed_label_config"],
                    ):
                        logging.warning(f"Found differences between expected and actual template: {differences}")
                return p
        return None

    def _create_new(self):
        assert self.template, "Template must be set to create a new project"
        p = self.client.start_project(
            title=f"{self.args.name}",
            label_config=self.template,
        )
        return p

    async def delele_all_predictions(self) -> list[int]:
        prediction_ids = [
            p["id"] for p in self.client.make_request("GET", f"/api/predictions?project={self.ls.id}").json()
        ]
        for pred_id in prediction_ids:
            self.client.make_request("DELETE", f"/api/predictions/{pred_id}")
        return prediction_ids

    def __repr__(self):
        return f"LSProject(name={self.args.name}, id={self.ls.id})"



from pathlib import Path
import logging
from functools import partial
import os
import pprint
import json
import hashlib

from typing import List, TypeVar, Type, Dict, Tuple, Literal
import numpy as np
import torch
import torch.nn.functional as nnF

from mmm.BaseModel import BaseModel

from shutil import rmtree
from pathlib import Path

from mmm.BaseModel import BaseModel
from mmm.logging.type_ext import StepMetricDict


def check_streamlit():
    """
    Function to check whether python code is run within streamlit or not.
    """
    try:
        import streamlit as st
        from streamlit.runtime.scriptrunner import get_script_run_ctx

        if not get_script_run_ctx():
            use_streamlit = False
        else:
            use_streamlit = True
    except ModuleNotFoundError:
        use_streamlit = False
    return use_streamlit


def convert_tile_into_patchbatch(region: torch.Tensor, patch_size=224, stride=224) -> Tuple[torch.Tensor, int, int]:
    """
    Converts a large tensor into a batch of smaller patches.

    If the region is not divisable by the stride, it will be padded with zeros *in-place*.

    That means that the input tensor might have a different shape after applying this function.

    Returns:
        torch.Tensor: A tensor of shape [B, C, patch_size, patch_size] where B is the number of patches.
        int: The number of patches in the width dimension (columns).
        int: The number of patches in the height dimension (rows).

    >>> import torch
    >>> from mmm.utils import convert_tile_into_patchbatch
    >>> testtensor = torch.Tensor(
    ...     [
    ...         [1, 1],
    ...         [1, 1],
    ...         [2, 2],
    ...         [2, 2],
    ...         [3, 3],
    ...         [3, 3],
    ...     ]
    ... )
    >>> patches, c, r = convert_tile_into_patchbatch(testtensor.unsqueeze(0), patch_size=2, stride=2)
    >>> assert False not in (patches[0, 1] == torch.Tensor([[2., 2.],[2., 2.]]))
    >>> c
    3
    >>> r
    1

    """

    region = region.unsqueeze(0)

    # Calculate padding size for height and width
    # h, w = region.shape[-2:]
    # pad_h = (stride - h % stride) % stride
    # pad_w = (stride - w % stride) % stride
    # assert pad_h == 0 and pad_w == 0
    # if pad_h > 0 or pad_w > 0:
    #     region = nnF.pad(region, (0, pad_w, 0, pad_h), mode='constant', value=0)

    # torch.Size([1, C, patches_by_height(rows), patches_by_width(columns), patch_size, patch_size])
    # patches = region.unfold(2, patch_size, stride).unfold(3, patch_size, stride)
    patches = region.unfold(2, patch_size, stride).unfold(3, patch_size, stride)
    # return patches
    rows, columns = patches.shape[2], patches.shape[3]

    # Transform into a batch of patches [B, C, patch_size, patch_size]
    patchbatch = patches.reshape(1, 3, -1, patch_size, patch_size).squeeze(dim=0)
    # patchbatch
    return patchbatch.permute(1, 0, 2, 3), rows, columns


def make_divisable_by(x: int, by: int = 32):
    """Increases number to be divisable by `by`"""
    return x + ((by - x) % by)


def unique_str_hash(*args, **kwargs):
    """
    Currently, the order of keyword argument matter
    """
    argumenttuple = tuple(args) + tuple([x for keyvalue in kwargs.items() for x in keyvalue])
    argumenthasher = hashlib.md5()
    for arg in argumenttuple:
        # Different types might need to be hashed differently
        if isinstance(arg, Path):
            argumenthasher.update(arg.absolute().__str__().encode())
        else:
            argumenthasher.update(str(arg).encode())
    return argumenthasher.hexdigest()


def get_default_cachepath(folder_name: str = "default", based_on="ML_DATA_CACHE") -> Path:
    if based_on not in os.environ:
        # "Set the cache_path manually or set it using the {based_on} env var"
        cache_path = Path("~").expanduser() / ".mmm"
    else:
        cache_path = Path(os.getenv(based_on, default=f"./{folder_name}"))
    res = cache_path / folder_name
    res.mkdir(parents=True, exist_ok=True)
    return res


def disk_cacher(cache_path: Path | Literal["local", "shared"] = "local", disable: bool = False):
    """
    Can cache the result of pure functions.
    In other words, the result needs to be unique w.r.t. the unique combination of function name and its parameters.

    If cache_path is "local", then the local cache path by convention will be used.
    If cache_path is "shared", then the shared cache path by convention will be used.

    It will put a hash.pkl file with the pickled function result in the cache path.
    The hash results from the stringified parameters and the stringified values of the keyword parameters.

    If its fine for you to cache into RAM, use `functools.lru_cache` instead.
    """
    if cache_path == "local":
        cache_path = get_default_cachepath("pure_functions", based_on="ML_DATA_CACHE")
    elif cache_path == "shared":
        cache_path = get_default_cachepath("cache/pure_functions", based_on="ML_DATA_OUTPUT")

    def function_wrapper(func):
        def wrapper(attempt: int, *args, **kwargs):
            function_cache_dir = cache_path / func.__name__
            if not function_cache_dir.exists():
                function_cache_dir.mkdir(parents=True)

            # If args already known: return cached result
            func_hash = unique_str_hash(*args, **kwargs)
            p = function_cache_dir / f"{func_hash}.pkl"
            if p.exists() and not disable:
                try:
                    with open(p, "rb") as f:
                        return torch.load(f)
                except EOFError as e:
                    # This cache seems to be invalid, remove it!
                    logging.warn(f"Removing and recomputing {p} because of exception {e}, {attempt=}")
                    os.remove(p)
                    # And compute it again
                    if attempt > 0:
                        raise Exception(f"Using cache failed with {e} for {p=} {args=} {kwargs=}")
                    else:
                        return wrapper(attempt + 1, *args, **kwargs)

            # Else: execute function and return cache its result:
            function_result = func(*args, **kwargs)
            # and save
            with open(p, "wb") as f:
                torch.save(function_result, f)
            return function_result

        return partial(wrapper, 0)

    return function_wrapper


def remove_folder_blocking_if_exists(folder: Path):
    if folder.exists():
        rmtree(folder)
        while folder.exists():
            pass  # Wait for rmtree to finish cleaning up


T = TypeVar("T", bound=BaseModel)


def load_config_from_str(basemodel: Type[T], config_str: str, verbose=True) -> T:
    if config_str:
        logging.debug(config_str)
        config = basemodel(**json.loads(config_str))
    else:
        config = basemodel()

    if verbose:
        # Only import if user wants to know the difference of the config to its default!
        from deepdiff import DeepDiff

        differences = DeepDiff(basemodel().dict(), config.dict())
        if differences:
            diff_dict = differences.to_dict()
            print(f"Using adapted default config with {list(diff_dict.keys())}! (see logging.debug for details)")
            logging.debug(diff_dict)
        else:
            print(f"Using default config!")

    return config


def load_config_from_json5(basemodel: Type[T], file_path: Path, verbose=True) -> T:
    import json5

    with open(file_path, "r") as f:
        return load_config_from_str(basemodel, json.dumps(json5.load(f)), verbose=verbose)


def load_config_from_env(basemodel: Type[T], config_env_name: str, verbose=True) -> T:
    """
    If an environment variable with name `config_env_name` exists,
    will return the basemodel with adaptions made from the json-content of that environment variable.
    Otherwise, returns the default model.

    If `verbose`, will print the differences to the base configuration.
    """
    if config_env_name in list(os.environ.keys()):
        json_string = os.getenv(config_env_name)
        assert json_string is not None, f"{config_env_name} environment variable was not found"
    else:
        json_string = ""
    return load_config_from_str(basemodel, json_string, verbose=verbose)


def flatten_list_of_dicts(step_metrics: List[StepMetricDict]) -> StepMetricDict:
    """
    Concatenates arrays for every top-level key. For example,

    >>> import numpy as np
    >>> from mmm.utils import flatten_list_of_dicts
    >>> flatten_list_of_dicts([
    ...     {"a": np.array([1, 2]), "b": np.array([1, 2])},
    ...     {"a": np.array([3, 4])}
    ... ])
    {'a': array([1, 2, 3, 4]), 'b': array([1, 2])}
    """
    metric_arrs: Dict[str, List[np.ndarray]] = {}
    for d in step_metrics:
        for k, v in d.items():
            if k not in metric_arrs:
                metric_arrs[k] = []

            metric_arrs[k].append(v)

    metrics = {k: np.concatenate(arrs) for k, arrs in metric_arrs.items()}
    return metrics


def recursive_equality(e1, e2, approx=False) -> bool:
    """
    Compares objects. Supports torch.Tensor.
    """
    if type(e1) is not type(e2):
        return False

    # If e1 has a __len__ method, the lengths of e1 and e2 must be equal
    try:
        if len(e1) != len(e2):
            return False
    except TypeError:
        pass

    if isinstance(e1, dict):
        if set(e1.keys()) != set(e2.keys()):
            return False
        # Assumes equal order
        return False not in [recursive_equality(v1, v2) for v1, v2 in zip(e1.values(), e2.values())]
    elif isinstance(e1, list) or isinstance(e1, tuple):
        return False not in [recursive_equality(x1, x2) for x1, x2, in zip(e1, e2)]
    elif isinstance(e1, torch.Tensor):
        if approx:
            return torch.allclose(e1, e2)
        else:
            return torch.equal(e1, e2)
    elif isinstance(e1, np.ndarray):
        if approx:
            return np.allclose(e1, e2)
        else:
            return np.array_equal(e1, e2)
    elif e1 != e1:
        # Only true for nan values
        return np.isnan(e1) and np.isnan(e2)
    else:
        return e1 == e2


# Plotting


def create_sample_efficiency_plot_from_wandb(
    run_id: str,
    project_name: str,
    downstream_tasks: List,
    entity: str = "tissue-concepts",
    metric: str = "acc",
):
    """
    Function to create a sample-efficiency plot between different downstream tasks

    run_id: id of the wandb run to import data from
    project_name: project in which the run is stored

    downstream_tasks: ['str', 'str'] list fo tasks. Example: ["breakhis", "bach"]
    entitiy: wandb entitiy. Ususally tissue-concepts
    metric: metric to be plottet at y-axis

    return: plotly figure object
    """
    import wandb
    import re
    import plotly.graph_objects as go

    api = wandb.Api()
    run = api.run(f"{entity}/{project_name}/{run_id}/")

    summary = run.summary
    summary_keys = list(summary.keys())

    fig = go.Figure()

    for task in downstream_tasks:
        r = re.compile(f"epoch_tl.bestval(.*){task}(.*)_{metric}")
        found = list(filter(r.match, summary_keys))
        data = [[int(entry.split(f"{task}")[-1].split("_")[0]), summary[entry]] for entry in found]

        data.sort()
        x = [a[0] for a in data]
        y = [a[1] for a in data]

        fig.add_trace(go.Scatter(x=x, y=y, name=task))

    fig.update_layout(xaxis_title="Percent Samples", yaxis_title=f"best overall {metric}")

    return fig


def find_missing_torchrun_envvars() -> list[str]:
    torchrun_envvars = [
        "LOCAL_RANK",
        "RANK",
        "WORLD_SIZE",
        "LOCAL_WORLD_SIZE",
        "TORCHELASTIC_RESTART_COUNT",
        "TORCHELASTIC_MAX_RESTARTS",
        "TORCHELASTIC_RUN_ID",
    ]
    return [var for var in torchrun_envvars if var not in os.environ]



from __future__ import annotations
import json
from ast import Not
from typing import Any, List, Dict, Mapping, Tuple, Optional, Type, Literal
from typing_extensions import Annotated
import random
from PIL.Image import Image as PIL_Image
import logging

import wandb
import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import Dataset

from pydantic import Field

from mmm.logging.type_ext import StepMetricDict
from mmm.logging.wandb_ext import build_wandb_image_for_clf

from .MTLTask import MTLTask
from mmm.settings import mtl_settings
from mmm.data_loading.TrainValCohort import TrainValCohort
from mmm.data_loading.ClassificationDataset import ClassificationDataset
from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.mtl_modules.shared_blocks.Grouper import Grouper, make_grid_for_supercase
from mmm.neural import LossConfigs, CrossEntropyLossConfig
from mmm.mtl_modules.shared_blocks.SharedModules import SharedModules
from mmm.neural.pooling import GlobalPooling, GlobalPoolingConfig
from mmm.mtl_modules.shared_blocks.Grouper import Grouper

from mmm.utils import flatten_list_of_dicts
from sklearn.metrics import (
    roc_auc_score,
    accuracy_score,
    f1_score,
    top_k_accuracy_score,
)


class ClassificationTask(MTLTask):
    """
    Special case of a MTLTask which deals with classification directly from the backbone's output $$Z$$.

    It requires cohorts holding classification datasets.

    Labels are encoded by their index in `class_names`.
    In consequence, the order needs to correspond to the labels!!
    """

    class Config(MTLTask.Config):
        encoder_key: str = "encoder"
        squeezer_key: str = "squeezer"
        grouper_key: str = Field(
            default="",
            description="If set, assumes a grouper to exist in the shared modules.",
        )
        loss_fn: Annotated[LossConfigs, Field(discriminator="loss_type")] = CrossEntropyLossConfig()
        dropout: float = 0.2
        metrics: Optional[List[Literal["confusion matrix", "accuracy", "top5accuracy", "auc", "f1"]]] = Field(
            default=None,
            description="If none, the task will decide which metrics make sense",
        )
        head: Literal["pretraining", "smart"] = "pretraining"

    @classmethod
    def from_torchvision_style(
        cls: Type[ClassificationTask],
        hidden_dim: int,
        class_names: List[str],
        task_config: Config,
        cohort_config: TrainValCohort.Config,
        train_ds: Dataset[Tuple[PIL_Image, torch.Tensor]],
        val_ds: Dataset[Tuple[PIL_Image, torch.Tensor]],
    ) -> ClassificationTask:
        """
        Automatically builds a task from torchvision default datasets.

        Torchvision datasets provide their training examples using tuples (PIL_Image, class_index: Integer-torch-tensor)
        """
        cohort = TrainValCohort(
            cohort_config,
            ClassificationDataset.from_torchvision(train_ds, class_names=class_names),
            # If no validation set is specified, the trainer will use cross-validation from the training set
            ClassificationDataset.from_torchvision(val_ds, class_names=class_names),  # if val_ds is not None else None
        )
        return cls(hidden_dim, class_names, task_config, cohort)

    def __init__(
        self,
        hidden_dim: int,
        args: Config,
        cohort: TrainValCohort[ClassificationDataset],
    ):
        super().__init__(args, cohort)
        self.args: ClassificationTask.Config  # Make sure IDE knows about the task specific fields
        self.class_names = cohort.datasets[0].get_classes_for_visualization()
        self.hidden_dim = hidden_dim

        if self.args.head == "pretraining":
            self.task_modules = self._create_pretraining_head()
        else:
            self.task_modules = self._create_smart_head()

        self._grouper_weights = None
        self.flatten = nn.Flatten(1)
        self.criterion: nn.Module = self.args.loss_fn.build_instance()

    def _create_pretraining_head(self):
        out_dim = len(self.class_names)
        new_dict = nn.ModuleDict(
            {
                "classification_head": nn.Sequential(
                    nn.Dropout(p=self.args.dropout),
                    nn.ReLU(),
                    nn.Linear(self.hidden_dim, out_dim),
                )
            }
        )
        return new_dict

    def _create_smart_head(self) -> nn.ModuleDict:
        out_dim = len(self.class_names)
        new_dict = nn.ModuleDict(
            {
                "classification_head": nn.Sequential(
                    nn.Dropout(p=self.args.dropout),
                    nn.ReLU(),
                    nn.Linear(
                        max(out_dim * 4, self.hidden_dim),
                        max(out_dim * 4, self.hidden_dim // 2),
                    ),
                    nn.Dropout(p=self.args.dropout),
                    nn.ReLU(),
                    nn.Linear(self.hidden_dim // 2, out_dim),
                )
            }
        )
        return new_dict

    def prepare_batch(self, batch: Dict[str, Any]) -> Any:
        batch["image"] = batch["image"].to(self.torch_device)
        batch["class"] = batch["class"].to(self.torch_device)
        return batch

    def forward(self, inputs, shared_blocks: Dict[str, SharedBlock]):
        x, supercase_indexes = inputs
        pyr = shared_blocks[self.args.encoder_key](x)
        _, hidden_vector = shared_blocks[self.args.squeezer_key](pyr)
        hidden_vector = self.flatten(hidden_vector)

        if self.args.grouper_key:
            hidden_vector, self._grouper_weights = shared_blocks[self.args.grouper_key](
                hidden_vector, supercase_indexes
            )

        out = self.task_modules["classification_head"](hidden_vector)
        return out

    def training_step(self, batch: Dict[str, Any], shared_blocks: SharedModules):
        x = batch["image"]
        y = batch["class"]

        # skip if batch is empty
        tensornums = [bool(t.numel()) for t in x]
        if not True in tensornums:
            logging.info("encountered batch without valid training examples, skipping batch")
            return None

        # If a grouper is used, extract supercase_indices
        if self.args.grouper_key:
            # A batch with ids ["id1", "s3", "s3"] would become [0, 1, 1]
            grouper: Grouper = shared_blocks.module.shared_modules[self.args.grouper_key]
            supercase_indices = grouper.extract_ids_from_batch([x["group_id"] for x in batch["meta"]]).to(
                self.torch_device
            )

            # batch sizes might be unified here for some settings of the Grouper.
            # x = grouper.unify_bagsizes(bag=x, supercase_indices=supercase_indices)

            # the targets need to be grouped as well, currently y is a (B,) tensor with class indices
            # For each unique supercase index, we need to find the corresponding class index
            y = grouper.group_targets(y, supercase_indices)
        else:
            supercase_indices = None

        y_hat = shared_blocks.forward((x, supercase_indices), self.forward)
        batch_loss = self.criterion(y_hat, y) / np.log(len(self.class_names))

        step_results: StepMetricDict = {  # type: ignore (.numpy() does not correctly indicate numpy array)
            "targets": y.cpu().numpy(),
            "logits": y_hat.detach().cpu().numpy(),
            "preds": torch.argmax(y_hat.detach().cpu(), dim=1).numpy(),
        }
        self.add_step_result(batch_loss.item(), step_results)

        live_vis = self._visualize_preds(
            x.detach().cpu(),
            step_results,
            (batch["meta"] if "meta" in batch else [{} for _ in range(batch["image"].shape[0])]),
            supercase_indices=supercase_indices,
        )

        return batch_loss, live_vis

    def _visualize_preds(
        self, training_ims, step_metrics: dict, metas: list[dict], supercase_indices
    ) -> Dict[str, Any]:
        vis_n = min(self._takeout_vis_budget(), training_ims.size(0))

        if vis_n <= 0:
            return {}

        if supercase_indices is not None:
            # Select one of the groups for visualization
            group_index = random.choice(list(set(supercase_indices.cpu().numpy())))
            grid_img, weight_str, vis_indices = make_grid_for_supercase(
                training_ims, supercase_indices, group_index, self._grouper_weights
            )

            caption = f"""
Group {group_index} with {len(vis_indices)} subcases, group id: {metas[0]["group_id"]}
weights:
{weight_str}
logits:
{step_metrics["logits"][group_index]}
{[metas[i] for i in vis_indices]=}
            """
            wandb_img, description, true_str, pred_str = build_wandb_image_for_clf(
                grid_img,
                step_metrics["targets"][group_index],
                step_metrics["preds"][group_index],
                self.class_names,
                caption_suffix=caption,
            )
            return {"preds": [wandb_img]}
        else:
            preds = []
            for rand_index in random.sample(list(range(training_ims.size(0))), vis_n):
                metastr = json.dumps(metas[rand_index], default=lambda o: str(o))
                wandb_img, description, true_str, pred_str = build_wandb_image_for_clf(
                    training_ims[rand_index],
                    step_metrics["targets"][rand_index],
                    step_metrics["preds"][rand_index],
                    self.class_names,
                    caption_suffix=metastr,
                )
                preds.append(wandb_img)
            return {"preds": preds} if preds else {}

    def _get_short_class_names(self, max_length=10):
        if True in [len(c) > max_length for c in self.class_names]:
            return [f"{i};{c[:max_length]}" for i, c in enumerate(self.class_names)]
        else:
            return self.class_names

    def log_epoch_metrics(self) -> Tuple[Dict[str, Any], str]:
        metrics = flatten_list_of_dicts(self._step_metrics)
        if self.args.metrics is None:
            if len(self.class_names) <= mtl_settings.max_classes_detailed_logging:
                selected_metrics = ["confusion matrix", "accuracy", "auc", "f1"]
            else:
                selected_metrics = ["accuracy", "top5accuracy"]
        else:
            selected_metrics = self.args.metrics

        _, print_str = super().log_epoch_metrics()
        log_dict = {}

        if "accuracy" in selected_metrics:
            log_dict["acc"] = accuracy_score(y_true=metrics["targets"], y_pred=metrics["preds"])
            print_str = f"{print_str} - acc: {log_dict['acc']}"

        if "top5accuracy" in selected_metrics:
            classes_in_loop = np.unique(metrics["targets"])
            classes_in_loop.sort()
            log_dict["top5acc"] = top_k_accuracy_score(
                y_true=metrics["targets"],
                y_score=metrics["logits"][:, classes_in_loop],
                k=5,
                labels=classes_in_loop,
            )
            print_str = f"{print_str} - top5acc: {log_dict['top5acc']}"

        if "auc" in selected_metrics:
            # Binary classification
            if metrics["logits"].shape[1] == 2:
                preds = nn.Softmax(dim=1)(torch.from_numpy(metrics["logits"]))[:, 1]
            else:
                preds = nn.Softmax(dim=1)(torch.from_numpy(metrics["logits"]))

            try:
                log_dict["auc"] = roc_auc_score(metrics["targets"], preds, multi_class="ovr")

                print_str = f"{print_str} - auc: {log_dict['auc']}"
            except ValueError as e:
                logging.warn(f"Computing auc failed with {e}")

        if "confusion matrix" in selected_metrics:
            log_dict["confmat"] = wandb.plot.confusion_matrix(
                preds=metrics["preds"],  # type: ignore
                y_true=metrics["targets"],  # type: ignore
                class_names=self._get_short_class_names(),
                title=f"{self._prefix}_{self.get_name()}",
            )

        if "f1" in selected_metrics:
            log_dict["f1"] = f1_score(y_true=metrics["targets"], y_pred=metrics["preds"], average="macro")

        return log_dict, print_str



from typing import Union
from .TorchModule import TorchModule
from .modules import *
from .losses import (
    CrossEntropyLossConfig,
    MSELossConfig,
    RMSELossConfig,
    FocalLossConfig,
    FocalLoss,
    NLLLossConfig,
)
from .activations import ActivationFunctionConfig, ActivationFn

LossConfigs = Union[
    CrossEntropyLossConfig,
    MSELossConfig,
    RMSELossConfig,
    FocalLossConfig,
    NLLLossConfig,
]



from __future__ import annotations
import warnings
from tqdm.auto import tqdm
import numpy as np
from pydantic import Field
from typing import Optional, Literal, Iterable
from mmm.BaseModel import BaseModel
from typing import Optional, Tuple, Dict, Any, Generic, TypeVar, Callable
import logging

import torch
import torch.multiprocessing as mp
from torch.utils.data import DataLoader

from mmm.utils import unique_str_hash, get_default_cachepath
from mmm.DataSplit import DataSplit
from .MTLDataset import MTLDataset, DatasetStyle

DatasetType = TypeVar("DatasetType", bound=MTLDataset, covariant=True)


class TrainValCohort(Generic[DatasetType]):
    """
    Can be used to define a training cohort with a training and a validation dataset.
    It is used to create dataloaders.

    By setting cache-foldername you allow the cohort to cache its datasets.
    The cached values will appear in ML_DATA_CACHE/cache_key/*.pkl
    """

    class Config(BaseModel):
        batch_size: tuple[int | None, int | None] = Field(default=(None, 1))
        shuffle_loaders: tuple[bool, bool] = (True, True)
        num_workers: int = Field(
            default=1,
            description="""
Number of workers that prepare batches for this task.
Training loader and validation loader will ask for num_workers worker.
Should be chosen such that the task can run alone without any bottlenecks.
In case too many workers are requested the trainer will reduce the number of workers across all tasks
while keeping a minimum of one worker per task.
""",
        )
        pin_memory: bool = False
        prefill_cache: bool = True

    def __init__(
        self,
        args: Config,
        train_ds: DatasetType,
        val_ds: DatasetType,
        cache_foldername="",
    ) -> None:
        self.args: TrainValCohort.Config = args
        self.datasets: Tuple[DatasetType, DatasetType] = (train_ds, val_ds)
        self.data_loaders: Tuple[Optional[DataLoader], Optional[DataLoader]] = (
            None,
            None,
        )

        if cache_foldername:
            ignore_keys_for_discoverability = [
                "batch_size",
                "pin_memory",
                "shuffle_loaders",
                "num_workers",
                "prefill_cache",
            ]  # , "cross_val_split_size"]
            reidentification_args = self.args.dict().copy()
            for key in ignore_keys_for_discoverability:
                reidentification_args.pop(key)
            logging.info(f"I think {cache_foldername} cache needs to be unique w.r.t. \n {reidentification_args}")
            unique_config_hash = unique_str_hash(**reidentification_args)
            self.cache_path = get_default_cachepath(folder_name=cache_foldername) / unique_config_hash
            self.datasets[0].enable_caching(self.cache_path / "train")
            self.datasets[1].enable_caching(self.cache_path / "val")
        else:
            self.cache_path = None

    def __repr_html__(self) -> str:
        return f"""
        <pre><code>{self.__repr__()}</pre></code>
        """

    def _st_repr_(self) -> None:
        from mmm.logging.st_ext import stw, st

        # Streamlit would otherwise resample cross validation splits
        import streamlit as st

        # with st.():
        if st.sidebar.button(f"Run `prepare_epoch(epoch=0)`"):
            self.prepare_epoch(epoch=0)

        if split_name_selection := st.sidebar.selectbox("TrainOrVal", ["Training", "Validation"]):
            split_name: str = split_name_selection
        else:
            split_name: str = "Training"
        train_val_index = 0 if split_name == "Training" else 1

        stw(self.datasets[train_val_index])

    def __repr__(self) -> str:
        return (
            f"Cohort with args: {self.args}"
            f"\nTrain dataset: {self.datasets[0]}\nValidation dataset: {self.datasets[1]}"
        )

    def get_random_batch(self, split: DataSplit) -> Dict[str, Any]:
        warnings.warn(
            "get_random_batch is deprecated. Use ds.get_random_batch instead.",
            DeprecationWarning,
        )
        if None in self.data_loaders:
            self.prepare_epoch(epoch=0)

        return next(iter(self.data_loaders[split.value]))  # type: ignore

    def get_dataloader(self, data_split: DataSplit) -> DataLoader:
        dl = self.data_loaders[data_split.value]
        assert dl is not None, f"Prepare dataloader for {data_split} first using cohort.prepare_epoch"
        return dl

    def build_iterator(self, data_split: DataSplit) -> Iterable:
        # Map-style datasets might have a different length every time
        if self.datasets[data_split.value].get_dataset_style() == DatasetStyle.MapStyle:
            cur_len = len(self.datasets[data_split.value].src_ds)
            if cur_len > len(self.datasets[data_split.value]) * self.datasets[data_split.value].reduced_size:
                logging.debug(
                    f"Dataset {data_split} of {self} has grown from {len(self.datasets[data_split.value])} to {cur_len}."
                )
                # Due to multiprocessing the worker keeps a copy of the old, shorter dataset. It needs to be terminated.
                self.terminate_datasplit_workers(data_split)
                # Let the MTLDataset know about the new length
                self.datasets[data_split.value].reset_indices()

        return iter(self.data_loaders[data_split.value])

    def get_dataset(self, data_split: DataSplit) -> MTLDataset:
        ds = self.datasets[data_split.value]
        assert ds is not None
        return ds

    def get_active_workers(self, group: DataSplit):
        if self.data_loaders[group.value] is not None and self.data_loaders[group.value]._iterator is not None:
            return self.args.num_workers
        else:
            return 0

    def terminate_datasplit_workers(self, group: DataSplit):
        if self.data_loaders[group.value] is not None and self.data_loaders[group.value]._iterator is not None:
            self.data_loaders[group.value]._iterator._shutdown_workers()  # type: ignore
            self.data_loaders[group.value]._iterator = None

    def terminate_workers(self):
        """
        Terminating workers loses the state of the current iterator.
        As a result, you should never do this during usage (e.g. a loop).
        """
        self.terminate_datasplit_workers(DataSplit.train)
        self.terminate_datasplit_workers(DataSplit.val)

    def prepare_epoch(self, epoch: int):
        """
        The epoch might be used by child classes to seed splits for cross validation.
        """
        # Caches do not need to be specific to temporary cross validation datasets
        # because augmentations are not cached anyways
        if (self.cache_path is not None) and self.args.prefill_cache:
            self.datasets[0].prefill_cache()
            self.datasets[1].prefill_cache()

        # If deterministic datasets are used, then dataloaders do not need to be renewed:
        if None in self.data_loaders:
            assert None not in self.args.batch_size, f"Batch size must be set for creating dataloaders, {epoch=}"
            train_loader = self.datasets[0].get_dataloader(
                shuffle=self.args.shuffle_loaders[0],
                batch_size=self.args.batch_size[0],
                num_workers=self.args.num_workers,
                pin_memory=self.args.pin_memory,
                # Usually, this only improves performance, but it is important for federated learning.
                persistent_workers=self.args.num_workers > 0,
            )
            val_loader = self.datasets[1].get_dataloader(
                shuffle=self.args.shuffle_loaders[1],
                batch_size=self.args.batch_size[1],
                num_workers=self.args.num_workers,
                pin_memory=self.args.pin_memory,
                # Usually, this only improves performance, but it is important for federated learning.
                persistent_workers=self.args.num_workers > 0,
            )

            self.data_loaders = train_loader, val_loader

    def get_onnx_input(self, device: str):
        """
        Exporting to ONNX requires an example batch.

        Currently, it only works for the classification task.
        """
        if self.data_loaders:
            self.prepare_epoch(epoch=0)

        loader_index = 0
        assert self.data_loaders[loader_index] is not None, "Dataloader shouldn't be None here"
        example_batch = next(iter(self.data_loaders[loader_index]))  # type: ignore

        # Specific to classification
        example_input = example_batch["image"].to(device)

        return example_input

    def transform_cohort_to_sklearn(
        self,
        feature_encoder: Callable,
        get_untransformed_cases: bool = True,
        include_meta: bool = False,
        batch_size: int = 1,
        num_samples: int = -1,
    ):
        """
        If your feature encoder function is a shared block, you can use the `block.torch_device` property
        for finding out the device of the shared block.

        The dataset's method `get_input_output_tuple` is used for transforming the batch to sklearn's format.
        """

        def _transform_dataloader(ds: MTLDataset):
            x_batches = []
            y_batches = []
            meta = []

            active_children_before = len(mp.active_children())
            dl = ds.get_mp_batchiterator(
                get_untransformed_cases=get_untransformed_cases,
                batch_size=batch_size,
                persistent_workers=True,
            )
            for i, one_batch in enumerate(tqdm(dl)):
                if include_meta:
                    meta.append(torch.Tensor(one_batch["meta"]))
                x_raw, y = self.datasets[0].get_input_output_tuple(one_batch)
                x = feature_encoder(x_raw)
                x_batches.append(x)
                y_batches.append(y)
                if num_samples > 0 and i > num_samples:
                    break

            dl._iterator._shutdown_workers()  # type: ignore
            dl._iterator = None

            if len(mp.active_children()) > active_children_before:
                logging.warn(f"After killing workers, there are still {len(mp.active_children())} workers")

            if dl._iterator is not None:
                logging.warn(f"had to manually shutdown workers")
                dl._iterator._shutdown_workers()  # type: ignore
                dl._iterator = None
            if not meta:
                meta = torch.Tensor([])
            else:
                meta = torch.concat(meta)
            return torch.concat(x_batches).cpu(), torch.concat(y_batches), meta

        X_train, y_train, meta_train = _transform_dataloader(self.datasets[0])
        X_val, y_val, meta_val = _transform_dataloader(self.datasets[1])
        return X_train, y_train, X_val, y_val, meta_train, meta_val



from abc import abstractmethod
from mmm.BaseModel import BaseModel
import torch.nn as nn


class TorchModule(BaseModel):
    """
    Baseclass for pydantic models which can be directly instantiated into a torch.nn.Module
    """

    @abstractmethod
    def build_instance(self, *args, **kwargs) -> nn.Module:
        raise NotImplementedError



from __future__ import annotations
import logging
import urllib
import os
from fsspec.core import OpenFile
import fsspec
from upath import UPath
from pydantic import Field, BaseModel, model_validator

import torch.multiprocessing as mp
from torch.utils.data import get_worker_info

# Gets the worker info for the parent process
worker_info = get_worker_info()
mp_method = mp.get_start_method()


class DistributedPath(BaseModel):
    """
    Wraps fsspec with a pydantic model:

    >>> from mmm.data_loading.DistributedPath import DistributedPath
    >>> dp = DistributedPath(uri="s3://dataroot/histoartefact_preannotation/SynD_0_001.svs")
    >>> fs, f = dp.fs(), dp.file()

    Can be used with paths like "s3://bucketname/file", "/path/to/file" and other fsspec URIs.
    If no options are provided, it will default to Minio settings: S3URL, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY.

    Use `upath` for a pathlib like interface provided by `universal_pathlib` package:

    >>> assert len(list(dp.upath().parent.glob("*"))) > 0

    """

    uri: str = Field(..., description="FSSpec URI to the file")
    options: dict = Field({}, description="Options to pass to fsspec")

    @model_validator(mode="before")
    @classmethod
    def serialize_from_string(cls, data):
        if isinstance(data, str):
            data = DistributedPath(uri=data).model_dump()

        return data

    @model_validator(mode="after")
    def fill_s3_defaults(self):
        if self.uri.startswith("s3://") and not self.options:
            self.options = {
                "endpoint_url": os.getenv("S3URL", default="http://localhost:9000"),
                "key": os.getenv("AWS_ACCESS_KEY_ID", default="minioadmin"),
                "secret": os.getenv("AWS_SECRET_ACCESS_KEY", default="minioadmin"),
            }

        return self

    def fix_multiprocessing_with_fork(self):
        global worker_info
        if mp_method == "fork" and (new_worker_info := get_worker_info()) is not None:
            # If the worker changed, some stuff will deadlock if not erased now:
            if worker_info is None:
                logging.info(f"MP {mp_method=}, erasing parent fsspec thread")
                fsspec.asyn.iothread[0] = None
                fsspec.asyn.loop[0] = None
                worker_info = new_worker_info

    def file(self, *args, **kwargs) -> OpenFile:
        self.fix_multiprocessing_with_fork()
        return fsspec.open(self.uri, *args, **kwargs, **self.options)

    def upath(self, *args, **kwargs) -> UPath:
        self.fix_multiprocessing_with_fork()
        return UPath(self.uri, *args, **kwargs, **self.options)

    def fs(self):
        self.fix_multiprocessing_with_fork()
        return fsspec.filesystem(self.get_protocol(self.uri), **self.options)

    def get_protocol(self, path):
        # Parse the path using urllib.parse
        parsed_path = urllib.parse.urlparse(path)

        # Extract the scheme (protocol) from the parsed path
        return parsed_path.scheme if parsed_path.scheme else "file"

    def exists(self) -> bool:
        return self.fs().exists(self.uri)

    def __truediv__(self, other):
        new_uri = os.path.join(self.uri, other)
        return DistributedPath(uri=new_uri, options=self.options)



"""
This module is part of the shared blocks, because it implements knowledge about an active task.
As a consequence, it is not a generally reusable PyTorch nn.Module.
"""

from copy import deepcopy

import torch.nn as nn


class TaskSpecificLayer(nn.Module):
    """
    Converts a layer into a task specific layer.

    A layer for a new task is copied from the `self.original_layer`.

    If you do not call `create_layer_for_task`, the original layer is used for your forward pass.
    In this case, an exception is raised.
    """

    def __init__(self, layer: nn.Module):
        super().__init__()
        self.task_modules: nn.ModuleDict = nn.ModuleDict({})
        self.active_task = ""
        self.original_layer = layer

    def create_layer_for_task(self, task_id: str) -> None:
        if task_id not in self.task_modules:
            self.task_modules[task_id] = deepcopy(self.original_layer)
        else:
            raise Exception(f"Task {task_id} already known to this task-specific layer")

    def set_active_task(self, task_id: str):
        self.active_task = task_id

    def forward(self, *args, **kwargs):
        if self.active_task in self.task_modules:
            return self.task_modules[self.active_task].forward(*args, **kwargs)
        else:
            if self.active_task != "original":
                raise Exception(
                    f"Using forward pass of original layer in {self}. "
                    + f"Because of unknown active task {self.active_task}"
                    + f"To suppress this warning set activate task to 'original'"
                )
            return self.original_layer(*args, **kwargs)



import os
import typer

app = typer.Typer()


@app.command()
def status():
    from mmm.settings import mtl_settings

    typer.secho("MMM Settings", fg=typer.colors.BRIGHT_BLUE)
    typer.secho(mtl_settings.model_dump_json(indent=2))

    # Test if mmm is importable
    typer.secho("Testing if mmm is importable...", fg=typer.colors.BRIGHT_BLUE)
    from mmm.interactive import blocks, configs, data, pipes, tasks, training, utils

    typer.secho("mmm is importable!", fg=typer.colors.GREEN)


@app.command()
def inference_app(host: str = "0.0.0.0", port: int = 8000):
    try:
        import uvicorn
        import fastapi
    except ImportError:
        typer.secho("Please install the extra dependencies 'api' of medicalmultitaskmodeling", fg=typer.colors.RED)
        return
    from mmm.inference_app import build_app, APISettings

    typer.secho(
        f"Use environment variables to configure the app: {APISettings.Config.env_prefix}", fg=typer.colors.BRIGHT_BLUE
    )
    if "MTLAPI_modules_path" not in os.environ:
        typer.secho(
            f"For example: `MTLAPI_modules_path=/path/to/model.pt.zip mmm inference-app` to use local model",
            fg=typer.colors.BRIGHT_BLUE,
        )
    api_settings = APISettings()
    app = build_app(api_settings)

    uvicorn.run(app, host=host, port=port)


if __name__ == "__main__":
    app()



"""
Configurable datatypes for selecting iterations in a loop for something to happen.
"""

from abc import abstractmethod
from typing import List, Literal, Union
from mmm.BaseModel import BaseModel
from pydantic import validator

import random
import math


class EventSelectorBase(BaseModel):
    @abstractmethod
    def is_event(self, iteration: int) -> bool:
        raise NotImplementedError()


class FixedEventSelector(EventSelectorBase):
    selector_type: Literal["fixed"] = "fixed"
    at_iterations: List[int] = [0]

    def is_event(self, iteration: int):
        return iteration in self.at_iterations


class RecurringEventSelector(EventSelectorBase):
    selector_type: Literal["recurring"] = "recurring"
    every_n: int = 1
    starting_at: int = 0

    def is_event(self, iteration: int):
        return iteration % self.every_n == 0 and iteration >= self.starting_at


class CodedEventSelector(EventSelectorBase):
    """Enables code injection, do not allow in apps which use external configs!"""

    selector_type: Literal["coded"] = "coded"
    python_program: str = r"i % random.randint(2, 8) == 0"

    @validator("python_program")
    def must_be_valid_program(cls, _):
        res = cls._execute_prog(0)
        assert isinstance(res, bool), "program must execute to bool"

    def _execute_prog(self, iteration: int):
        return eval(self.python_program, {"i": iteration, "random": random, "math": math})

    def is_event(self, iteration: int):
        return self._execute_prog(iteration)


class CombinedEventSelector(EventSelectorBase):
    selector_type: Literal["combined"] = "combined"
    events: List[Union[FixedEventSelector, RecurringEventSelector, CodedEventSelector]]

    def is_event(self, iteration: int):
        subevent_results = [v.is_event(iteration) for v in self.events]
        return True in subevent_results


EventSelector = Union[
    FixedEventSelector,
    RecurringEventSelector,
    CodedEventSelector,
    CombinedEventSelector,
]



"""
Task samplers are used to build multi-task optimization steps.

Task samplers mainly require a cohort. However, they require a list of tasks because tasks have an id.
Further, a single task might reference multiple cohorts in the future.
"""

from __future__ import annotations

import itertools
from typing import Optional, List, Any, Tuple, Generator, Sized, cast, Dict
import random
import logging
from abc import ABC, abstractmethod
from typing import Literal, Union

from torch.utils.data import DataLoader

from mmm.BaseModel import BaseModel
from mmm.data_loading.TrainValCohort import TrainValCohort
from mmm.data_loading.MTLDataset import MTLDataset, DatasetStyle
from mmm.mtl_modules.tasks.MTLTask import MTLTask
from mmm.DataSplit import DataSplit


class BaseSampler(ABC):
    class Config(BaseModel):
        sampler_type: str

    def __init__(self, args: Config, tasks: List[MTLTask], loader_index: DataSplit) -> None:
        """
        loader_index DataSplit.train for the training data loader, DataSplit.val for validation
        """
        self.args = args
        self.loader_index: DataSplit = loader_index
        self.tasks: List[MTLTask] = tasks
        task_ids = [t.get_name() for t in self.tasks]
        assert len(task_ids) == len(set(task_ids))

        self._iterator = None

    def _get_dataset_styles(self) -> List[DatasetStyle]:
        dss: List[MTLDataset] = [t.cohort.get_dataset(self.loader_index) for t in self.tasks]
        return [ds.get_dataset_style() for ds in dss]

    @abstractmethod
    def is_finite(self) -> bool:
        """
        A finite task sampler has a __len__ property
        """
        raise NotImplementedError()

    def __len__(self) -> int:
        """
        The total number of batches of all tasks combined. Only finite task samplers have a length.
        """
        raise NotImplementedError(f"Task sampler {self} does not have a length")

    @abstractmethod
    def _get_generator(self) -> Generator:
        raise NotImplementedError(f"Task sampler {self} needs to implement __iter__ protocol.")

    def continue_iter(self):
        if self._iterator is None:
            self._iterator = self._get_generator()
        return self._iterator

    def __iter__(self):
        return self._get_generator()


class ConcatTaskSampler(BaseSampler):
    """
    Iterates over all tasks and yields each batch one after another.
    All batches of task n will be returned before the first batch of task n+1.

    All cohort's dataloaders need to have a finite length.
    """

    class Config(BaseSampler.Config):
        sampler_type: Literal["concat"] = "concat"

    def __init__(self, args: Config, tasks: List[MTLTask], loader_index: DataSplit):
        super().__init__(args, tasks, loader_index)
        self.loader_lengths: List[int] = [len(t.cohort.get_dataloader(self.loader_index)) for t in self.tasks]

    def is_finite(self) -> bool:
        return True

    def __len__(self) -> int:
        return sum(self.loader_lengths)

    def _get_generator(self):
        for task in self.tasks:
            task_loader: DataLoader = task.cohort.get_dataloader(self.loader_index)
            for x in task_loader:
                yield x, task
            logging.debug(f"Done iterating through {task.get_name()}")


class BalancedTaskSampler(BaseSampler):
    """
    Randomly yields tasks, such that all tasks are drained exactly once.

    Weights are only computed in the beginning,
    as a result extra care might have to be taken with map-style datasets that change their length over time.
    """

    class Config(BaseSampler.Config):
        sampler_type: Literal["balanced"] = "balanced"

    def __init__(self, args: Config, tasks: List[MTLTask], loader_index: DataSplit):
        super().__init__(args, tasks, loader_index)

        def get_weight(cohort: TrainValCohort) -> float:
            if False not in [s is DatasetStyle.MapStyle for s in self._get_dataset_styles()]:
                return len(cohort.get_dataloader(self.loader_index)) / self.__len__()
            else:  # if the dataloader doesn't have a fixed len, approximate!
                return 1.0 / len(self.tasks)

        self.task_weights = [(t, get_weight(t.cohort)) for t in self.tasks]
        logging.debug(f"{self.task_weights=}")
        self.iters = {t.get_name(): t.cohort.build_iterator(self.loader_index) for t in self.tasks}

    def __len__(self) -> int:
        return sum([len(t.cohort.get_dataloader(self.loader_index)) for t in self.tasks])

    def is_finite(self) -> bool:
        return False not in [s is DatasetStyle.MapStyle for s in self._get_dataset_styles()]

    def _get_generator(self):
        while len(self.task_weights) > 0:
            t, task_weight = random.choices(self.task_weights, weights=[t[1] for t in self.task_weights], k=1)[0]
            try:
                yield next(self.iters[t.get_name()]), t
            except StopIteration:
                self.task_weights.remove((t, task_weight))
                logging.debug(f"Removing {t.get_name()} from task sampler")


class CyclicTaskSampler(BaseSampler):
    """
    Alternates deterministically through the tasks.

    If `break_with_shortest_loader` is True, `len(shortest_dataloader) * len(tasks)` is the length.
    If not, the task sampler is infinite.
    """

    class Config(BaseSampler.Config):
        sampler_type: Literal["cyclic"] = "cyclic"
        mode: Literal[
            "break_with_shortest_loader", "break_with_longest_loader", "infinite"
        ] = "break_with_longest_loader"

    def __init__(self, args: Config, tasks: List[MTLTask], loader_index: DataSplit):
        self.args: CyclicTaskSampler.Config
        super().__init__(args, tasks, loader_index)

    def __len__(self) -> int:
        assert self.is_finite()

        if self.args.mode == "break_with_shortest_loader":
            relevant_length = min(
                [
                    len(t.cohort.get_dataloader(self.loader_index))  # type: ignore (len of "possibly" None)
                    for t in self.tasks
                ]
            )
        else:
            relevant_length = max(
                [
                    len(t.cohort.get_dataloader(self.loader_index))  # type: ignore (len of "possibly" None)
                    for t in self.tasks
                ]
            )
        return relevant_length * len(self.tasks)

    def is_finite(self) -> bool:
        styles: List[DatasetStyle] = self._get_dataset_styles()
        if DatasetStyle.IterStyle in styles:
            return False
        else:
            if False in [s is DatasetStyle.MapStyle for s in styles]:
                # Once a weird thing happenened and if does happen again, this warning exposes it
                logging.warning([s is DatasetStyle.MapStyle for s in styles])
            assert DatasetStyle.IterStyle not in styles
            return self.args.mode in [
                "break_with_shortest_loader",
                "break_with_longest_loader",
            ]

    def _get_generator(self) -> Generator[Tuple[Any, MTLTask], None, None]:
        # iter_task_tuples = [(iter(t.cohort.get_dataloader(self.loader_index)), t)
        #                     for t in self.tasks]
        iter_task_tuples = [(t.cohort.build_iterator(self.loader_index), t) for t in self.tasks]
        if self.is_finite():
            counting_iter = range(self.__len__() // len(iter_task_tuples))
        else:
            counting_iter = itertools.count()

        tasks_that_stopped, renew_and_continue = [], True

        for _ in counting_iter:
            for task_index, (batch_iterator, task) in enumerate(iter_task_tuples):
                task: MTLTask
                try:
                    yield next(batch_iterator), task
                except StopIteration:
                    if task.get_name() not in tasks_that_stopped:
                        tasks_that_stopped.append(task.get_name())

                    if self.args.mode == "break_with_shortest_loader":
                        assert len(tasks_that_stopped) == 1
                        logging.debug(f"Breaking loop with task {task.get_name()} because it was shortest")
                        renew_and_continue = False
                    elif self.args.mode == "break_with_longest_loader":
                        renew_and_continue = len(tasks_that_stopped) < len(self.tasks)
                        logging.debug(f"{tasks_that_stopped=}, {renew_and_continue=}, {len(self.tasks)=}")
                    elif self.args.mode == "infinite":
                        renew_and_continue = True
                    else:
                        raise Exception(f"Unknown mode {self.args.mode}")
                    if renew_and_continue:
                        logging.debug(f"RENEWING ITERATOR FOR TASK {task.get_name()}")
                        # ite = iter(task.cohort.get_dataloader(self.loader_index))
                        ite = task.cohort.build_iterator(self.loader_index)
                        yield next(ite), task
                        iter_task_tuples[task_index] = (ite, task)
            if not renew_and_continue:
                break


TaskSamplerTypes: Dict = {
    ConcatTaskSampler.Config().sampler_type: ConcatTaskSampler,
    BalancedTaskSampler.Config().sampler_type: BalancedTaskSampler,
    CyclicTaskSampler.Config().sampler_type: CyclicTaskSampler,
}
TaskSamplerConfig = Union[
    ConcatTaskSampler.Config,
    BalancedTaskSampler.Config,
    CyclicTaskSampler.Config,
]



import logging
from typing import Dict
import torch.nn as nn
import torch.distributed as dist

from .SharedBlock import SharedBlock


# class Dummy(nn.Module):
#     def __init__(self):
#         super().__init__()

#     def forward(self, x, shared_modules):
#         return x


class SharedModules(nn.Module):
    """
    Intended to be wrapped by DDP to avoid a varying number of forward passes,
    leading to blocking.
    """

    def __init__(self, shared_blocks: Dict[str, SharedBlock]):
        super().__init__()
        self.shared_modules = nn.ModuleDict(shared_blocks)

        # if not dist.is_initialized():
        #     # If we are not using DDP, make the shared modules available in the same way as DDP
        #     self.module = Dummy()
        #     self.module.shared_modules = self.shared_modules

    @property
    def module(self):
        return self

    def forward(self, x, f):
        # logging.info(f"!!!forward{f.__name__}!!!")
        return f(x, self.shared_modules)



from abc import abstractmethod
from pathlib import Path
from typing import List, Literal, Mapping, Sequence, Tuple
import io

try:
    import onnx
except ImportError:
    pass
import torch
import torch.nn as nn

from mmm.mtl_modules.MTLModule import MTLModule
from mmm.torch_ext import replace_childen_recursive
from mmm.neural.module_conversions import (
    convert_2d_to_3d,
    build_instancenorm_like,
)

from mmm.mtl_modules.MTLModule import MTLModule
from .TaskSpecificLayer import TaskSpecificLayer

ModelInput = torch.Tensor | List[torch.Tensor]


class SharedBlock(MTLModule):
    """
    A shared block is intended to contain mainly parameters shared among multiple tasks.
    A shared block can have task-specific layers, new tasks can be introduced by `prepare_for_task(...)`

    A classic example is an encoder which encodes an image into a disentangled latent space.
    """

    class Config(MTLModule.Config):
        norm_layer: Literal[
            "nonorm",
            "instancenorm",
            "keepnorm",
            "layernorm",
            "taskspecific-batchnorm",
            "affinelayernorm",
        ] = "affinelayernorm"

    def __init__(self, args: Config) -> None:
        super().__init__(args)
        self.args = args
        self.next_active_task: str = ""
        self.task_specific_modules: List[TaskSpecificLayer] = []
        self._made_mtl_compatible = False

    def make_mtl_compatible(self):
        # Set normalization layer
        if self.args.norm_layer == "keepnorm":
            # shared blocks should use batchnorm as default
            pass
        elif self.args.norm_layer == "layernorm" or self.args.norm_layer == "affinelayernorm":
            if self.args.norm_layer == "layernorm":
                # Transformer models come with Layernorms by default, those only need to lose their learnable params
                replace_childen_recursive(
                    self,
                    nn.LayerNorm,
                    lambda oldnorm: nn.LayerNorm(
                        normalized_shape=oldnorm.normalized_shape,
                        eps=oldnorm.eps,
                        elementwise_affine=False,
                    ),
                )
            # CNNs come with Batchnorms by default, those can be replaced by Groupnorm with one group
            replace_childen_recursive(
                self,
                nn.modules.batchnorm._NormBase,
                lambda oldnorm: nn.GroupNorm(
                    1,
                    oldnorm.num_features,
                    affine=self.args.norm_layer == "affinelayernorm",
                ),
            )
        elif self.args.norm_layer == "instancenorm":
            replace_childen_recursive(self, nn.modules.batchnorm._NormBase, build_instancenorm_like)
        elif self.args.norm_layer == "nonorm":
            replace_childen_recursive(self, nn.modules.batchnorm._NormBase, lambda _: nn.Identity())
        elif self.args.norm_layer == "taskspecific-batchnorm":
            # Batchnorm supertype taken from the PyTorch implementation of SyncBatchNorm.convert_sync_batchnorm
            self._make_layertype_task_specific(nn.modules.batchnorm._BatchNorm)
        else:
            raise Exception(f"Unknown normlayer {self.args.norm_layer}")

        self._made_mtl_compatible = True

    def convert_to_3d(self):
        raise DeprecationWarning("Converting to 3D encoder is currently untested")
        for k, layer in self.named_children():
            setattr(self, k, convert_2d_to_3d(layer))

    def _make_layertype_task_specific(self, layertype_to_replace):
        """
        Replaces all layers of a certain type by a wrapper module which copies that layer for each task.
        """

        def task_specific_constructor(oldlayer) -> TaskSpecificLayer:
            newlayer = TaskSpecificLayer(oldlayer)
            self.task_specific_modules.append(newlayer)
            return newlayer

        replace_childen_recursive(self, layertype_to_replace, task_specific_constructor)

    def prepare_for_task(self, task_id: str) -> None:
        """
        Needs to be called once before training for each task if task-specific layers are used.
        """
        for l in self.task_specific_modules:
            l.create_layer_for_task(task_id)

    def set_active_task(self, task_id: str):
        """
        Needs to be called once before each forward pass if task-specific layers are used.
        """
        for l in self.task_specific_modules:
            l.set_active_task(task_id)

    def get_example_input(self) -> ModelInput | Tuple[ModelInput, ...]:
        """
        Used for example in tracing for ONNX export.
        """
        return torch.rand(1, 3, 224, 224).to(self.torch_device)

    def get_input_names(self) -> Sequence[str]:
        return ["input"]

    def get_output_names(self) -> Sequence[str]:
        return ["output"]

    def get_dynamic_axes(self) -> Mapping[str, Mapping[int, str]]:
        return {}

    def export_to_onnx(self, path: Path, for_task: str = "") -> None:
        if onnx is None:
            raise ImportError("ONNX not installed, cannot export model")

        self.set_active_task(for_task)

        example_input = self.get_example_input()
        example_output_shapes = [
            output_tensor.shape
            for example_output in (self(*example_input) if isinstance(example_input, tuple) else self(example_input))
            for output_tensor in (example_output if isinstance(example_output, list) else (example_output,))
        ]
        output_names = self.get_output_names()
        dynamic_axes = self.get_dynamic_axes()

        with io.BytesIO() as buffer:
            torch.onnx.export(
                self.eval(),  # model being run
                example_input,  # model input (or a tuple for multiple inputs)
                buffer,  # where to save the model (can be a file or file-like object)
                export_params=True,  # store the trained parameter weights inside the model file
                opset_version=13,  # the ONNX version to export the model to
                do_constant_folding=True,  # whether to execute constant folding for optimization
                input_names=self.get_input_names(),  # the model's input names
                output_names=output_names,  # the model's output names
                training=torch.onnx.TrainingMode.EVAL,
                dynamic_axes=self.get_dynamic_axes(),
            )
            buffer.seek(0)
            onnx_model = onnx.load(buffer)

        # Set the correct model output shapes since shape inference fails for some of our models
        for i, (name, shape) in enumerate(zip(output_names, example_output_shapes)):
            for d, s in enumerate(shape):
                if (name not in dynamic_axes or d not in dynamic_axes[name]) and onnx_model.graph.output[
                    i
                ].type.tensor_type.shape.dim[d].dim_param:
                    onnx_model.graph.output[i].type.tensor_type.shape.dim[d].dim_param = ""
                    onnx_model.graph.output[i].type.tensor_type.shape.dim[d].dim_value = s

        onnx.save_model(onnx_model, str(path))

    @abstractmethod
    def forward(self, input):  # type: ignore (weird override thing from PyTorch)
        raise NotImplementedError



"""
Collection of Cohorts and Datasets for Self-supervised Tasks
"""

from __future__ import annotations

import torch
from typing import Any, Callable, Dict, List, Iterator, Optional
from torch.utils.data import Dataset
import torchvision.transforms as transforms
import random
import numpy as np
import PIL

from torch.utils.data import Dataset, IterableDataset
from mmm.transforms import ApplyToKey
from mmm.logging.st_ext import side_by_side
from mmm.data_loading.MTLDataset import MTLDataset
from mmm.data_loading.ClassificationDataset import ClassificationDataset
from mmm.data_loading.TrainValCohort import TrainValCohort
from mmm.torch_ext import combine_datasets
from mmm.augmentations import SimCLRPatchAug, get_contrastive_2D_augs


class SSLDSWrapper(IterableDataset):
    def __init__(self, src) -> None:
        super().__init__()
        self.src = src
        self.src.batch_transform = None
        self.to_tensor = transforms.ToTensor()

    def channel_first(self, img: np.ndarray) -> np.ndarray:
        shape = img.shape
        # print(shape)
        if shape[0] == 3:
            return img
        elif shape[2] == 3:
            return np.transpose(img, (2, 1, 0))
        elif shape[1] == 3:
            return np.transpose(img, (1, 0, 2))
        else:
            channel_idx = np.argmin(shape)
            if shape[channel_idx] == 1:
                if shape[0] == 1:
                    return np.stack([img, img, img], axis=0)
                else:
                    img = np.stack([img, img, img], axis=2).squeeze()
                    return np.transpose(img, (2, 1, 0))
            else:
                raise Exception()

    def __iter__(self):
        """
        The price you pay for different Dataloding strategies
        """
        for item in self.src:
            if isinstance(item, Dict):
                img = item["image"]
            elif isinstance(item, tuple):
                img = item[0]
            else:
                continue
            if isinstance(img, torch.Tensor):
                img = img.numpy()
            elif isinstance(img, PIL.Image.Image):
                img = np.asarray(img)
            else:
                continue
            if img.max() > 1.0:
                img = img / 255
            # img = torch.from_numpy(self.channel_first(img.squeeze()))
            img = torch.from_numpy(img)
            yield {"image": img.float(), "target": img.float()}


class SSLDataset(MTLDataset):
    def __init__(
        self,
        src_ds: Dataset,
        src_transform: Optional[Callable],
        input_transform: Optional[Callable],
        target_transform: Optional[Callable],
    ) -> None:
        super().__init__(
            src_ds,
            src_transform=None,
            optional_case_keys=["meta"],
            mandatory_case_keys=["image", "target"],
        )
        self.src_ds: Dataset = src_ds
        self.src_transform = src_transform
        self.input_transform = input_transform
        self.target_transform = target_transform

    def set_classes_for_visualization(self, classes: List[str]):
        self.vis_classes = [str(i) for i in range(len(self.src_ds))]

    def __getitem__(self, index: int) -> Dict[str, Any]:
        d = super().__getitem__(index)

        if self.src_transform is not None:
            d = self.src_transform(d)

        if self.target_transform is not None:
            d["target"] = self.target_transform(d["target"])

        if self.input_transform is not None:
            d["image"] = self.input_transform(d["image"])

        return d

    def __iter__(self) -> Iterator[Dict[str, Any]]:
        for d in iter(self.src_ds):
            if self.src_transform is not None:
                d = self.src_transform(d)
            if self.input_transform is not None:
                d["image"] = self.input_transform(d["image"])
            if self.target_transform is not None:
                d["target"] = self.target_transform(d["target"])

            yield d

    def st_case_viewer(self, case: Dict[str, Any], index: int) -> None:
        import streamlit as st

        st.title("SSL view")
        side_by_side(case["image"], case["target"])
        st.write(case)


class AETrainValCohort(TrainValCohort[SSLDataset]):
    class Config(TrainValCohort.Config):
        shared_img_size: int = 224

    def __init__(self, args: Config, cohort_list) -> None:
        src_transform = transforms.Compose(
            [
                ApplyToKey(
                    transforms.Resize((args.shared_img_size, args.shared_img_size)),
                    key="image",
                ),
                ApplyToKey(
                    transforms.Resize((args.shared_img_size, args.shared_img_size)),
                    key="target",
                ),
            ]
        )

        train_ds = get_combined_SSL_dataset(list=[c.datasets[0] for c in cohort_list], src_transform=src_transform)

        val_ds = get_combined_SSL_dataset(
            list=[c.datasets[1] for c in cohort_list if c.datasets[1] is not None],
            src_transform=src_transform,
        )

        super().__init__(args, train_ds, val_ds)


class BTTrainValCohort(TrainValCohort[SSLDataset]):
    class Config(TrainValCohort.Config):
        shared_img_size: int = 224

    def __init__(self, args: Config, cohort_list) -> None:
        src_transform = transforms.Compose(
            [
                ApplyToKey(
                    transforms.Resize((args.shared_img_size, args.shared_img_size)),
                    key="image",
                ),
                ApplyToKey(
                    transforms.Resize((args.shared_img_size, args.shared_img_size)),
                    key="target",
                ),
            ]
        )

        train_ds = get_combined_SSL_dataset(
            list=[c.datasets[0] for c in cohort_list],
            src_transform=src_transform,
            input_transform=transforms.Compose(get_contrastive_2D_augs()),
            target_transform=transforms.Compose(get_contrastive_2D_augs()),
        )

        val_ds = get_combined_SSL_dataset(
            list=[c.datasets[1] for c in cohort_list if c.datasets[1] is not None],
            src_transform=src_transform,
            input_transform=transforms.Compose([transforms.RandomResizedCrop(224, scale=(0.08, 0.1))]),
            target_transform=transforms.Compose([transforms.RandomResizedCrop(224, scale=(0.08, 0.1))]),
        )

        super().__init__(args, train_ds, val_ds)


def get_combined_SSL_dataset(list: List[MTLDataset], src_transform, input_transform=None, target_transform=None):
    ds = combine_datasets(list)
    return SSLDataset(
        src_ds=SSLDSWrapper(ds),
        src_transform=src_transform,
        input_transform=input_transform,
        target_transform=target_transform,
    )


# def cl_from_cohort(clf_cohort: TrainValCohort,
#                    cl_func: Callable
#                    ) -> TrainValCohort[ContrastiveImagePairDataset]:
#     val_ds = ContrastiveImagePairDataset(
#         clf_cohort.datasets[1],
#         cl_func) if clf_cohort.datasets[1] is not None else None
#     cl_cohort = TrainValCohort(
#         clf_cohort.args,
#         train_ds=ContrastiveImagePairDataset(clf_cohort.datasets[0], cl_func),
#         val_ds=val_ds
#     )
#     if cl_cohort.datasets[0].get_dataset_style() == DatasetStyle.MapStyle:
#         cl_cohort.datasets[0].set_classes_for_visualization(
#             [f"case_{i}" for i in range(len(cl_cohort.datasets[0]))])
#         if cl_cohort.datasets[1] is not None:
#             cl_cohort.datasets[1].set_classes_for_visualization(
#                 [f"case_{i}" for i in range(len(cl_cohort.datasets[1]))])

#     return cl_cohort


if __name__ == "__main__":
    import os
    from mmm.logging.st_ext import multi_cohort_explorer
    import streamlit as st
    import pydantic
    from pathlib import Path
    from mmm.data_loading.medical.histo_patchclassification import (
        Kather100kClassificationCohort,
        Kather100kClassificationCohortConfig,
    )
    from mmm.data_loading.medical.conic import ConicSemSegTrainValCohort

    def create_ae_cohort(cfg, lst):
        return AETrainValCohort(cfg, lst)

    def create_bt_cohort(cfg, lst):
        return BTTrainValCohort(cfg, lst)

    root = Path(os.getenv("ML_HISTO_ROOT", default="/histo_root/"))
    kather_conic = create_ae_cohort(
        cfg=AETrainValCohort.Config(batch_size=10),
        lst=[
            Kather100kClassificationCohort(root, Kather100kClassificationCohortConfig()),
            ConicSemSegTrainValCohort(root, ConicSemSegTrainValCohort.Config()),
        ],
    )
    BT = create_bt_cohort(
        cfg=BTTrainValCohort.Config(batch_size=16, num_workers=2),
        lst=[
            Kather100kClassificationCohort(root, Kather100kClassificationCohortConfig()),
            ConicSemSegTrainValCohort(root, ConicSemSegTrainValCohort.Config()),
        ],
    )

    multi_cohort_explorer(
        {
            # "kather_conic": lambda: kather_conic ,
            "BT": lambda: BT
        }
    )



from functools import partial

import random
import json
from typing import Any, List, Optional, Callable, Dict
from jinja2 import Template
import re

from pathlib import Path
import tempfile
import imageio.v3 as imageio
import numpy as np
import torch
from torch.utils.data import Dataset

from mmm.mtl_types import RGBImage, MultiLabelSegmentation, AnnotatedImage
from mmm.typing_utils import get_colors
from mmm.data_loading.MTLDataset import MTLDataset, SrcCaseType
from mmm.typing_utils import rgbnumpy_to_base64
from mmm.labelstudio_ext.utils import (
    binary_mask_to_result,
    brush_annotation_to_npy,
)
import torchvision.transforms.functional as F

LS_TEMPLATE = r"""
<View>
<Image name="image" value="$image" zoom="true" zoomControl="true" rotateControl="false"/>
<HyperText name="Context" value="$htmlmeta" />
<BrushLabels name="segmentations" toName="image">
{% for class_name, color in classes.items() %}
<Label value="{{class_name}}" background="rgba({{color[0]}}, {{color[1]}}, {{color[2]}}, 0.2)"/>
{% endfor %}
</BrushLabels>
</View>
"""


def masks_from_ls(
    results: list[dict],
    shape: tuple[int, int],
    class_names: list[str],
    unlabeled_value=-1,
) -> np.ndarray:
    # Extract dict[classname, mask] from the annotation
    # if "annotations" in ls_task and len(ls_task["annotations"]) > 0:
    # anno = random.choice(ls_task["annotations"])
    masks_dict = {classname: mask for classname, mask in [brush_annotation_to_npy(result) for result in results]}
    if shape is None:
        shape = masks_dict[list(masks_dict.keys())[0]].shape
    masks = np.zeros([len(class_names), *shape], dtype=np.int64)
    labeled_pixels = np.zeros(shape, dtype=np.int64)

    for classname, mask in masks_dict.items():
        class_index = class_names.index(classname)
        labeled_pixels[mask > 0] += 1
        torchmask = (torch.from_numpy(mask) > 0).long()
        # torchmask[torchmask == 0] = unlabeled_value
        # The region is considered labeled if it is labeled with any class
        masks[class_index, ...] = torchmask
        # masks[~class_index, torchmask == 1] = 0

    # The region is considered unlabeled if it is unlabeled for all classes
    masks[:, labeled_pixels == 0] = unlabeled_value

    return masks


def multilabel_case_from_ls(npy_case: dict[str, Any]) -> dict[str, Any]:
    npy_case["image"] = F.to_tensor(npy_case["image"])
    if "masks" in npy_case:
        npy_case["masks"] = torch.from_numpy(npy_case["masks"])
    return npy_case


def ls_create_mlsemseg_template(class_names: list[str]) -> str:
    class_colors = get_colors(len(class_names))
    template = Template(LS_TEMPLATE)
    return template.render(
        classes={
            class_name: [int(r * 255), int(g * 255), int(b * 255)]
            for class_name, (r, g, b) in zip(class_names, class_colors)
        },
    )


def mtl_case_to_ls(mtl_case: Dict[str, Any], class_names: list[str], htmlmeta: str = "") -> Dict[str, Any]:
    """
    Labelstudio likes to get files in the form of links.

    This function converts the image to a base64 string by default.
    If you have a link, use the "imagelink" key to overwrite this behaviour.

    For example:
    {
        "imagelink": "https://picsum.photos/id/1/200/300",
        "masks": np.random.randint(0, 2, [2, 200, 300]),
    }
    """
    # Convert the [C, H, W] image to [H, W, C] numpy
    if "imagelink" not in mtl_case:
        npy_img = np.transpose(mtl_case["image"].numpy() * 255.0, [1, 2, 0]).astype(np.uint8)
        mtl_case["imagelink"] = rgbnumpy_to_base64(npy_img)

    if not htmlmeta and "meta" in mtl_case:
        htmlmeta = json.dumps(mtl_case["meta"], default=lambda o: str(o))

    return {
        "data": {
            "image": mtl_case["imagelink"],
            "htmlmeta": htmlmeta,
        },
        "annotations": [
            {
                "result": [
                    binary_mask_to_result(mtl_case["masks"][i].numpy(), class_names[i], "segmentations")
                    for i in range(len(mtl_case["masks"]))
                    if torch.max(mtl_case["masks"][i]).item() > 0
                ]
            }
        ],
    }


class MultiLabelSemSegDataset(MTLDataset):
    """
    Requires an image and a mask where each pixel can have zero or all classes.

    "image" shape: [C, H, W]
    "masks" shape: [#class_names, H, W]
    """

    def __init__(
        self,
        src_ds: Dataset[SrcCaseType],
        class_names: List[str] = None,
        *args,
        **kwargs,
    ) -> None:
        assert class_names is not None, "Segmentation dataset needs class names"
        self.class_names: List[str] = class_names
        super().__init__(src_ds, ["image", "masks"], ["original_size", "meta"], *args, **kwargs)

    def set_classes_for_visualization(self, classes: List[str]):
        self.class_names = classes

    def verify_case(self, d: Dict[str, Any]) -> None:
        super().verify_case(d)
        self.assert_image_data_assumptions(d["image"])
        # For example, MONAI tends to load some specific subclass of Tensor, exclude it!
        assert torch.is_tensor(d["masks"]), "mask should be of type tensor"
        assert isinstance(d["masks"], torch.LongTensor)
        assert list(d["masks"].shape) == [len(self.class_names)] + list(d["image"].shape[1:])
        assert torch.max(d["masks"]).item() <= 1, "Multilabel targets should be binary"
        assert len(d["image"].shape) == len(d["masks"].shape), "Each class should have its own channel"

    def st_case_viewer(self, case: Dict[str, Any], i: int = -1) -> None:
        from mmm.logging.st_ext import stw, st

        stw("### Untransformed image:")
        img, masks = case["image"], case["masks"]

        anno_img = AnnotatedImage(
            image=RGBImage(data=img),
            annotations=[MultiLabelSegmentation(data=masks, class_names=self.class_names)],
        )
        stw(anno_img, st_prefix=f"img{i}")
        stw(case, st_prefix=f"case{i}")

    def _compute_batchsize_from_batch(self, batch: Dict[str, Any]) -> int:
        return batch["image"].shape[0]

    def _visualize_batch_case(self, batch: Dict[str, Any], i: int) -> None:
        from mmm.logging.st_ext import stw, st

        if "meta" in batch:
            st.write(batch["meta"][i])
        stw(
            AnnotatedImage(
                image=RGBImage(data=batch["image"][i]),
                annotations=[MultiLabelSegmentation(data=batch["masks"][i], class_names=self.class_names)],
            ),
            st_prefix=f"b{i}",
        )
        # mask_uniques = torch.unique(patch_mask)
        # blend_with_mask(
        #     patch,
        #     patch_mask,
        #     caption_suffix=f"{i}/{self._compute_batchsize_from_batch(batch)}\nUniques: {mask_uniques}",
        #     classes=self.class_names,
        #     st_key=f"b{i}",
        # )

    def ls_create_template(self) -> str:
        return ls_create_mlsemseg_template(self.class_names)

    def ls_get_case(self, mtl_case: Dict[str, Any], htmlmeta: str = "") -> Dict[str, Any]:
        return mtl_case_to_ls(mtl_case, self.class_names, htmlmeta)

    @classmethod
    def from_ls(cls, ls_ds, ls_tpl: str, *args, **kwargs):
        """
        Creates a dataset from a labelstudio template.
        """
        class_names: list[str] = re.findall(r'<Label value="([^"]*)"', ls_tpl)
        return cls(
            src_ds=ls_ds,
            src_transform=cls.ls_src_transform(class_names),
            class_names=class_names,
            *args,
            **kwargs,
        )

    @staticmethod
    def ls_src_transform(class_names: list[str], unlabeled_value=-1) -> Dict[str, Any]:
        """
        Transforms a labelstudio case into an MTL case.

        If there are multiple annotations a random annotation is selected.
        This might be a problem if you use caching!
        """

        return partial(
            multilabel_case_from_ls,
            class_names=class_names,
            unlabeled_value=unlabeled_value,
        )



from typing import List, Tuple, Literal
import torch
import torch.nn as nn

from ..TorchModule import TorchModule
from ..model_protocols import EncoderModel
from mmm.BaseModel import BaseModel
from mmm.neural.activations import ActivationFn, ActivationFunctionConfig
from mmm.torch_ext import infer_stride_channels_from_features
from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock


class ConvNormActivation(nn.Module):
    """
    Order of operations is inspired by torchvision ResNet implementation.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        norm: bool,
        activation: ActivationFunctionConfig,
        conv_stride: int,
        padding=0,
        kernel_size=3,
    ) -> None:
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=conv_stride,
            padding=padding,
        )

        if norm:
            # Always preliminary define batch norm because shared blocks replace it with the configured norm.
            self.norm = nn.BatchNorm2d(out_channels)
        else:
            self.norm = nn.Identity()

        self.activation = activation.build_instance()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.activation(self.norm(self.conv(x)))


class MiniConvNet(nn.Module, EncoderModel):
    """
    Builds a feature pyramid by applying sequential convolution+activation blocks.

    Extracts the hidden vector using an Adaptivepoolingoperation and a Linear(deepest_depth, hidden_dim) layer.
    """

    class Config(BaseModel):
        architecture: Literal["miniconvnet"] = "miniconvnet"
        num_channels: int = 3
        num_filters: int = 32
        conv_stride: int = 2
        depth: int = 4
        norm: bool = True
        padding: int = 1
        activation: ActivationFunctionConfig = ActivationFunctionConfig(fn_type=ActivationFn.GeLU)

        def build_instance(self, *args, **kwargs):
            return MiniConvNet(self)

    def __init__(self, args: Config) -> None:
        super().__init__()
        self.args = args

        hidden_convs = [
            ConvNormActivation(
                self.args.num_filters * (i + 1) if i >= 0 else self.args.num_channels,
                self.args.num_filters * (i + 2),
                args.norm,
                args.activation,
                self.args.conv_stride,
                padding=self.args.padding,
            )
            for i in range(-1, self.args.depth - 1)
        ]
        self.cnn_layers = nn.ModuleList(hidden_convs)

        with torch.no_grad():
            self.out_channels, self.strides = infer_stride_channels_from_features(
                self.forward(torch.rand(1, self.args.num_channels, 224, 224))
            )

    def get_feature_pyramid_channels(self) -> List[int]:
        return self.out_channels

    def get_strides(self) -> List[int]:
        return self.strides

    def forward(self, x: torch.Tensor) -> List[torch.Tensor]:
        features = [x]
        for conv_layer in self.cnn_layers:
            features.append(conv_layer(features[-1]))
        return features


class MiniDecoder(SharedBlock):
    """
    Small decoder to be used as ImageGenerationDecoder during pre-training. Images will most likely look bad.
    Works on the lowest feature pyramid and doubles the size of the tensor with every level.

    >>> from mmm.neural.modules.simple_cnn import MiniDecoder
    >>> import torch
    >>> dec = MiniDecoder(MiniDecoder.Config(pixel_embedding=16, upsample_levels=2), [3, 32, 64, 128])
    >>> test = [torch.ones(2, 128, 5, 5)]
    >>> out = dec(test)
    >>> assert out.shape == torch.Size([2, 16, 20, 20])
    """

    class Config(SharedBlock.Config):
        module_name: str = "miniDecoder"
        pixel_embedding: int = 32
        activation: ActivationFunctionConfig = ActivationFunctionConfig(fn_type=ActivationFn.ReLU)
        upsample_levels: int = 4

    def __init__(
        self,
        args: Config,
        enc_out_channels: List[int],
    ):
        super().__init__(args)
        self.args: MiniDecoder.Config
        # blocks upsample such, that the output dim is about H//2, W//2 of the original input size
        hidden_convs = [
            [
                nn.ConvTranspose2d(
                    in_channels=(
                        enc_out_channels[-1] if i == args.upsample_levels else args.pixel_embedding * (i + 1) * 2
                    ),
                    out_channels=args.pixel_embedding * i * 2 if i > 1 else args.pixel_embedding,
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    output_padding=1,
                ),
                args.activation.build_instance(),
            ]
            for i in reversed(range(1, args.upsample_levels + 1))
        ]
        self.blocks = nn.ModuleList([item for sublist in hidden_convs for item in sublist])

        self.make_mtl_compatible()

    def get_output_dim_per_pixel(self) -> int:
        return self.args.pixel_embedding

    def get_upsampling_factor(self) -> int:
        return self.args.upsample_levels

    def forward(self, features: List[torch.Tensor]):
        feat = features[-1]
        for block in self.blocks:
            feat = block(feat)

        return feat



import random
from urllib.parse import urljoin
import pickle
import os
from pathlib import Path
import time
import requests
import numpy as np
from typing import Dict, Optional, Any, OrderedDict, Tuple, Literal
import imageio.v3 as imageio
from urllib3.exceptions import InsecureRequestWarning
from mmm.utils import get_default_cachepath, unique_str_hash

requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)  # type: ignore


# @disk_cacher()
def get_image_as_numpy(url: str, attempts=5) -> np.ndarray:
    res = requests.get(url, verify=os.getenv("NOMAD_CACERT", default=False))
    b = res.content
    for attempt in range(attempts):
        try:
            return imageio.imread(b).copy()
        except Exception as e:
            print(e)
            e1 = f"Error opening {url} as numpy image in attempt {attempt}"
            print(e1)
            e2 = f"Cannot convert {res.text} to an image with the following exception:"
            print(e2)
            time.sleep(np.random.randint(2, 59))

    # dump_f = os.getenv("ML_DATA_OUTPUT", default="/data_output/") + "wsi-service-dump.pkl"
    # with open(dump_f, 'wb') as f:
    #     torch.save(res.content, f)
    # logging.info(f"Dumped {res.text}")
    raise Exception(f"{attempts} failed attempts to open Image {url}")


async def get_image_as_numpy_async(url: str, session=None) -> np.ndarray:
    """
    Calls an http and returns the result as numpy.

    For example, for the full url to an endpoint `full_endpoint_url` outside of async you can use:

    ```python
    return_value: np.ndarray = await get_rest_endpoint_as_numpy(full_endpoint_url)
    ```
    """
    if session is None:
        import aiohttp

        async with aiohttp.ClientSession() as session:
            resp = await session.get(url)
    else:
        resp = await session.get(url)
    b = await resp.read()
    try:
        return imageio.imread(b).copy()
    except Exception as e:
        print(e)
        e1 = f"Error opening {url} as numpy image"
        print(e1)
        resp_text = await resp.text()
        e2 = f"Cannot convert {resp_text} to an image with the following exception:"
        print(e2)
        raise Exception(f"{e}\n{e1}\n{e1}")


def get_json(url: str) -> Any:
    return requests.get(url, verify=os.getenv("NOMAD_CACERT", default=False)).json()


class StorageMapper:
    def __init__(
        self,
        base_url: str,
        storage_path: Path,
        slideid_strategy: Literal["absolutepath", "filename"],
    ) -> None:
        """
        The storage path is necessary to know at which point to cut the paths to the slides.
        """
        self.base_url = base_url
        self.storage_path: Path = storage_path
        self.id_strategy = slideid_strategy

    def build_request_body(self, slide_id: str, absolute_slide_path: Path) -> dict:
        return {
            "slide_id": f"{slide_id}",
            "storage_type": "fs",
            "storage_addresses": [
                {
                    "address": str(absolute_slide_path.relative_to(self.storage_path)),
                    "slide_id": f"{slide_id}",
                    "main_address": True,
                    "storage_address_id": f"{slide_id}",
                }
            ],
        }

    def get_info(self, slide_id: str) -> dict:
        response = requests.get(f"{self.base_url}v3/slides/{slide_id}/")
        return response.json()

    def upload_slide(self, slide_id: str, slide_path: Path) -> str:
        request_body = self.build_request_body(slide_id, slide_path)
        response = requests.post(f"{self.base_url}v3/slides/", json=request_body)
        return response.json()

    def get_unique_slideid(self, absolute_slide_path: Path) -> str:
        if self.id_strategy == "absolutepath":
            relative_path = absolute_slide_path.relative_to(self.storage_path)
            return str(relative_path).replace("/", "_").replace(".", "_")
        elif self.id_strategy == "filename":
            return absolute_slide_path.name.replace(".", "_")


class WSIService:
    """
    Wrapper around one instance of a running WSI-service.

    Just instantiating the object will already make calls to the WSI-service for basic indexing stuff.
    """

    def __init__(self, base_url: str) -> None:
        self.base_url = base_url
        # Memorizes detailed info for each slide
        self._slideinfo_memoization: Dict[str, Any] = {}

    def get_is_alive_route(self) -> str:
        return urljoin(self.base_url, "alive")

    def get_slide_info(self, slide_id: str) -> Dict:
        if slide_id not in self._slideinfo_memoization:
            u = f"{self.base_url}v3/slides/{slide_id}/info?plugin=tiffslide"
            d = get_json(u)
            self._slideinfo_memoization[slide_id] = d

        return self._slideinfo_memoization[slide_id]

    def build_link_for_region(
        self,
        wsi_service_img_id: str,
        level=2,
        position: Optional[Tuple[int, ...]] = None,
        spatial_size: Optional[Tuple[int, ...]] = None,
        format="png",
        pos_level_0=False,
    ) -> str:
        u = f"{self.base_url}v3/slides/{wsi_service_img_id}/region/level/{level}"

        if position is None:
            position = (0, 0)
        if pos_level_0:
            position = (
                int(position[0] / self.get_slide_info(wsi_service_img_id)["levels"][level]["downsample_factor"]),
                int(position[1] / self.get_slide_info(wsi_service_img_id)["levels"][level]["downsample_factor"]),
            )
        u += f"/start/{'/'.join(map(str, position))}"

        if spatial_size is None:
            spatial_size = (
                self.get_slide_info(wsi_service_img_id)["levels"][level]["extent"]["x"] - position[0],
                self.get_slide_info(wsi_service_img_id)["levels"][level]["extent"]["y"] - position[1],
            )

        u += f"/size/{'/'.join(map(str, spatial_size))}"
        u += f"?image_format={format}"
        u += f"&plugin=tiffslide"
        return u


class LocalServiceMapper:
    """
    Uses an instance of WSI service and a StorageMapper to map local paths to WSI-service ids implicitly.
    If the mounts are correct this should feel local.

    The mounts are correct if both the WSI-service and the user have access to the same volume,
    which should be mounted into /data of the WSI-service.
    The storage mapper does not need access.

    It needs to memorize which slides are already indexed in the storage mapper.
    """

    def __init__(
        self,
        wsi_service_base: str,
        storage_mapper_base: str,
        storage_mapper_root: Path,
        slideid_strategy: Literal["absolutepath", "filename"],
    ) -> None:
        unique_filename = unique_str_hash(wsi_service_base, storage_mapper_base, storage_mapper_root)
        self.cache_file_path = get_default_cachepath() / unique_filename
        self.wsi_service = WSIService(wsi_service_base)
        self.storage_mapper = StorageMapper(storage_mapper_base, storage_mapper_root, slideid_strategy)

        if not self.cache_file_path.exists():
            # Create file
            self.cache_file_path.parent.mkdir(parents=True, exist_ok=True)
            with open(self.cache_file_path, "wb") as f:
                # Save using pickle
                pickle.dump({}, f)
        self._load_from_disk()

    def _load_from_disk(self):
        # Load cache file
        with open(self.cache_file_path, "rb") as f:
            self.slides: Dict[str, Path] = pickle.load(f)

    def get_slide_id_by_path(self, absolute_slide_path: Path) -> str:
        return self.storage_mapper.get_unique_slideid(absolute_slide_path)

    def index_wsis(self, relative_root, forget_old=False, wsi_glob: str = "**/*.svs"):
        if forget_old:
            self.slides = {}

        root = self.storage_mapper.storage_path / relative_root
        for i, wsi_path in enumerate(root.glob(wsi_glob)):
            slide_id = self.storage_mapper.get_unique_slideid(wsi_path)
            assert (not forget_old) or slide_id not in self.slides, f"Slide {slide_id} already indexed"
            if slide_id not in self.slides:
                self.storage_mapper.upload_slide(slide_id, wsi_path)
                self.slides[slide_id] = wsi_path

        # Commit the already indexed slide ids to the cache file
        with open(self.cache_file_path, "wb") as f:
            pickle.dump(self.slides, f)

    def get_viewer_url(self, slide_id: str) -> str:
        return urljoin(self.wsi_service.base_url, f"v3/slides/{slide_id}/viewer")

    def get_all_ids(self, implicitly_update=True) -> list[str]:
        if implicitly_update:
            self._load_from_disk()
        return list(self.slides.keys())

    def __repr__(self) -> str:
        return f"""
        LocalServiceMapper with {len(self.slides)} slides indexed.
        WSI-service address: {self.wsi_service.base_url}v3/docs/
        Storage mapper address: {self.storage_mapper.base_url}v3/docs
        """

    def select_random_window(self, slide_id: str, tile_size: int = 1024) -> tuple[str, tuple[int, int, int, int]]:
        """
        For the slide, returns a wsi-service link to a random window of size tile_size x tile_size.
        The second value is the region of interest.
        """
        img_info = self.wsi_service.get_slide_info(slide_id)
        # if "levels" not in img_info:
        #     raise Exception(f"Slide {slide_id} has no levels")
        num_levels = len(img_info["levels"])
        # Select a level, the higher the level the less probable it should be picked
        random_level = random.choices(range(num_levels), weights=[2**i for i in reversed(range(num_levels))])[0]
        downsample_fac = img_info["levels"][random_level]["downsample_factor"]
        dims = (
            img_info["levels"][random_level]["extent"]["x"],
            img_info["levels"][random_level]["extent"]["y"],
        )
        pos = random.randint(0, max(0, dims[0] - tile_size)), random.randint(0, max(0, dims[1] - tile_size))
        image_url = self.wsi_service.build_link_for_region(
            wsi_service_img_id=slide_id,
            level=random_level,
            position=pos,
            spatial_size=(tile_size, tile_size),
        )

        # Build HTML snippet for viewer
        roi_x0 = int(pos[0] * downsample_fac)
        roi_x1 = int(pos[0] * downsample_fac + (tile_size * downsample_fac))
        # openlayers has an inverted y axis
        roi_y0 = img_info["extent"]["y"] - int(pos[1] * downsample_fac)
        roi_y1 = img_info["extent"]["y"] - int(pos[1] * downsample_fac + (tile_size * downsample_fac))

        return image_url, (roi_x0, roi_y0, roi_x1, roi_y1), random_level



from typing import List, Literal
from pydantic import Field
import torch
import timm
from timm.layers.norm import LayerNorm2d
import torch.nn as nn

from mmm.torch_ext import replace_childen_recursive
from ..TorchModule import TorchModule
from ..model_protocols import EncoderModel


def replace_timmlayernorm2d_with_defaultcnnnorm(old_norm: LayerNorm2d):
    return nn.BatchNorm2d(num_features=old_norm.normalized_shape[0], eps=old_norm.eps)


valid_variants = [
    "convnext_femto",
    "convnext_pico",
    "convnext_nano",
    "convnext_tiny",
    "convnext_small",
    "convnext_large",
    "convnext_xlarge_384_in22ft1k",
]


class TimmEncoder(nn.Module, EncoderModel):
    """
    Valid variants have pretrained weights, can yield feature maps and have the correct stride.
    """

    class Config(TorchModule):
        architecture: Literal["timm"] = "timm"
        variant: Literal[tuple(valid_variants)] = "convnext_nano"  # type: ignore
        pretrained: bool = Field(default=False, description="Loads ImageNet1k weights")
        drop_rate: float = 0.0
        num_channels: int = 3

        def build_instance(self, *args, **kwargs) -> nn.Module:
            # Cannot be imported earlier due to CyclicImportError
            from mmm.neural.modules.TimmEncoder import TimmEncoder

            return TimmEncoder(self)

    def __init__(self, args: Config) -> None:
        super().__init__()
        self.args = args
        self.timm_model = timm.create_model(
            args.variant,
            pretrained=args.pretrained,
            drop_rate=args.drop_rate,
            features_only=True,
            in_chans=args.num_channels,
            # out_indices=(2, 3, 4)
            # drop_path_rate=0.2
            # global_pool=args.pooling
        )

        # Timm uses a hack which makes our automatic replacements fail
        replace_childen_recursive(self.timm_model, LayerNorm2d, replace_timmlayernorm2d_with_defaultcnnnorm)

    def get_feature_pyramid_channels(self) -> List[int]:
        return [self.args.num_channels] + [d["num_chs"] for d in self.timm_model.feature_info]

    def get_strides(self) -> List[int]:
        return [1] + self.timm_model.feature_info.reduction()

    def forward(self, input: torch.Tensor):
        feature_pyramid = [input] + self.timm_model.forward(input)
        return feature_pyramid



import time
import json
import logging
import requests
import datetime
import uuid
from typing import Optional, Any, TYPE_CHECKING
import requests
from urllib.parse import urlparse, urljoin
import os
from io import BytesIO
from PIL import Image, ImageOps
import imageio.v3 as imageio
from pathlib import Path
import tempfile
import numpy as np
import io
import base64

try:
    from label_studio_sdk import Client, Project
    from label_studio_converter.brush import (
        decode_rle,
        encode_rle,
        decode_from_annotation,
    )
except ImportError:
    if not TYPE_CHECKING:
        Client, Project = Any, Any

from mmm.typing_utils import get_colors
from mmm.utils import remove_folder_blocking_if_exists


def sync_to_disk(
    project: Project,
    data_folder: Path,
    annotations,
    case_writer,
    dateformat: str = r"%Y-%m-%dT%H:%M:%S.%fZ",
):
    """
    LabelStudio has this built in, just add a target storage
    """
    data_folder.mkdir(parents=True, exist_ok=True)
    tasks_updated_cachepath: Path = data_folder / "tasks_updated_at.json"
    if tasks_updated_cachepath.exists():
        tasks_updated_at = {
            int(annoid): datetime.datetime.strptime(taskupdated, dateformat)
            for annoid, taskupdated in json.loads(tasks_updated_cachepath.read_text()).items()
        }
    else:
        tasks_updated_at = {}
    # Contains a map from the unique labelstudio id to a updated_at date like: 2023-09-24T07:18:12.618719Z
    tasks_updated_at_new = {a["id"]: datetime.datetime.strptime(a["updated_at"], dateformat) for a in annotations}
    updated_tasks = {
        new_anno_id: newdate
        for new_anno_id, newdate in tasks_updated_at_new.items()
        if new_anno_id not in tasks_updated_at or tasks_updated_at[new_anno_id] < newdate
    }
    logging.info(f"Found tasks which need to be updated: {updated_tasks.keys()}")
    tasks_updated_at.update(updated_tasks)

    # Download all images and masks
    for taskdict in filter(lambda a: a["id"] in updated_tasks.keys(), annotations):
        casefolder = data_folder / f"{taskdict['id']}"
        logging.info(f"Updating case on disk for annotation {taskdict['id']} in {casefolder}")
        # clear if it already exists
        remove_folder_blocking_if_exists(casefolder)
        casefolder.mkdir(parents=True, exist_ok=True)
        case_writer(taskdict, casefolder)

    tasks_updated_cachepath.write_text(
        json.dumps(
            tasks_updated_at,
            indent=2,
            default=lambda x: datetime.datetime.strftime(x, dateformat),
        )
    )
    (data_folder / "mtlfiledataset.json").write_text(json.dumps(project.get_params(), indent=2))


def add_npimage_to_project(project: Project, img: np.ndarray, metainfo: Optional[dict] = None):
    """
    Writes the image to a temporary file
    """
    tmpfile = tempfile.NamedTemporaryFile(suffix=".jpg")
    imageio.imwrite(tmpfile.name, img)
    imported_ids = project.import_tasks(Path(tmpfile.name))
    if metainfo is None:
        metainfo = {}

    assert "shape" not in metainfo
    metainfo["shape"] = img.shape

    project.update_task(imported_ids[0], meta=metainfo)
    return imported_ids


def brush_annotation_to_npy(result) -> tuple[str, np.ndarray]:
    """
    Expects something like:

    {
      'original_width': 2678,
      'original_height': 2954,
      'image_rotation': 0,
      'value': {'format': 'rle',
        'rle': [],
        'brushlabels': ['lung']},
      'id': 'aaOhNtmESo',
      'from_name': 'tag',
      'to_name': 'image',
      'type': 'brushlabels',
      'origin': 'manual'
    }
    """
    width, height = result["original_width"], result["original_height"]
    classnames = result["value"]["brushlabels"]
    rle = decode_rle(result["value"]["rle"])
    # For unknown reasons, the rle is decoded into four channels
    mask_channels = np.reshape(rle, [height, width, 4])

    mask = mask_channels[:, :, 3]

    assert result["image_rotation"] == 0
    assert len(classnames) == 1
    return classnames[0], mask


def convert_task_to_seglabel(task, class_names: list[str], prefill_value: int = 0) -> np.ndarray:
    labels = [
        brush_annotation_to_npy(result)
        for annotation in task["annotations"]
        for result in annotation["result"]
        if result["type"] == "brushlabels" and result["value"]["brushlabels"][0] in class_names
    ]
    if not labels:
        return None

    # Combine the labels to a single mask
    mask = np.zeros(labels[0][1].shape, dtype=np.int64)
    if prefill_value != 0:
        mask.fill(prefill_value)
    for class_name, annotation_mask in labels:
        # Plus one, because otherwise its impossible to distinguish between background and foreground
        mask[annotation_mask > 0] = class_names.index(class_name)

    # Visualize the mask using
    # Image.fromarray(mask * 50).save("mask.png")
    return mask


def binary_mask_to_result(mask: np.ndarray, class_name: str, brush_name: str, score=None):
    mask_255 = (mask > 0).astype(np.uint8) * 255
    flat_mask = np.repeat(mask_255.ravel(), 4)
    rle = encode_rle(flat_mask)
    res = {
        "id": str(uuid.uuid4())[0:8],
        "type": "brushlabels",
        "value": {"rle": rle, "format": "rle", "brushlabels": [class_name]},
        "to_name": "image",
        "from_name": brush_name,
        "image_rotation": 0,
        "original_width": mask.shape[1],
        "original_height": mask.shape[0],
    }
    if score is not None:
        res["score"] = score
    return res


def mask_to_annotation(
    mask: np.ndarray, class_names: list[str], brush_names: list[str], ignore_index: int | None = 0
) -> list:
    """
    - mask: a numpy array with shape (H, W) containing the class indices

    project.create_annotation(
        task_id=project.tasks[-1]["id"],
        result=mask_to_annotation(mask, ["notannotated", "foreground"])
    )
    """
    return [
        binary_mask_to_result(mask == i, class_name, brush_name)
        for i, (class_name, brush_name) in enumerate(zip(class_names, brush_names))
        if i in mask and (ignore_index is None or i != ignore_index)
    ]


def download_image(url: str, attempts: int = 5, backoff: float = 1.0, ls_client=None, ensure_rgb=True) -> np.ndarray:
    if url.startswith("data:image"):
        bytestring = url
        img_bytes = io.BytesIO(base64.b64decode(bytestring.split(",")[1]))
        # np_img: np.ndarray = imageio.imread(img_bytes)
    elif not bool(urlparse(url).netloc):
        # relative URL, use ls_client
        assert ls_client is not None, f"Relative URL {url} but no ls_client provided"
        r = ls_client.make_request("GET", url)
        img_bytes = io.BytesIO(r.content)
        # np_img: np.ndarray = imageio.imread(r.content)
    else:
        try:
            # Otherwise, it should be a valid URL
            r = requests.get(url)
            img_bytes = io.BytesIO(r.content)
        except Exception as e:
            logging.debug(f"Problem opening {url} as numpy image in attempt {attempts}: {e}")
            if attempts == 0:
                logging.error(f"Error opening {url} as numpy image")
                logging.error(f"Cannot convert {url} to an image with the following exception: {e}")
                raise e
            else:
                time.sleep(backoff)
                return download_image(url, attempts - 1, backoff * 2)

    pil_image = Image.open(img_bytes)
    # Some images might have EXIF transformations
    pil_image = ImageOps.exif_transpose(pil_image)
    if ensure_rgb:
        pil_image = pil_image.convert("RGB")

    return np.array(pil_image)


def download_image_from_task(url: str, token: str, task: dict, image_name: str = "image") -> Image:
    headers = {"Authorization": f"Token {token}"}
    download_url = f"{url}{task['data'][image_name]}"
    response = requests.get(download_url, headers=headers)
    # Create PIL image from response
    image = Image.open(BytesIO(response.content))

    # Ensure the image is RGB, sometimes they come in grayscale. Check if there are 3 channels and if not, convert
    if image.mode != "RGB" or len(image.getbands()) != 3:
        image = image.convert("RGB")

    return image



import torch.nn as nn
from typing import List, Any, get_type_hints, Dict, Callable, Type, Optional, TypeVar
from inspect import signature


def base_norm_replacement(oldnorm: nn.Module, replacements: Dict):
    for old_type, new_type in replacements.items():
        if isinstance(oldnorm, old_type):
            return new_type(oldnorm.num_features)
    raise Exception(f"I do not know how to convert {type(oldnorm)} to instance norm")


def build_instancenorm_like(
    oldnorm: nn.Module,
) -> nn.modules.instancenorm._InstanceNorm:
    return base_norm_replacement(oldnorm, {nn.BatchNorm2d: nn.InstanceNorm2d, nn.BatchNorm3d: nn.InstanceNorm3d})


def replace_childen_recursive(m: nn.Module, layertype_to_replace, newlayer_constructor, suppress_inheritance=False):
    """
    newlayer_constructor gets an instance of the layer to be replaced.

    Use cases might be replacing all 2D layers by their respective 3D versions.
    """
    for k, layer in m.named_children():
        by_typecheck = suppress_inheritance and type(layer) is layertype_to_replace
        by_isinstancecheck = (not suppress_inheritance) and isinstance(layer, layertype_to_replace)
        # Replace the object in question
        if by_typecheck or by_isinstancecheck:
            if newlayer_constructor is None:
                raise Exception(f"Constructor of replacement of {layertype_to_replace} is None :(")
            newlayer = newlayer_constructor(layer)
            # Does not work for children's children, which is why this is a recursive function
            setattr(m, k, newlayer)

        replace_childen_recursive(
            layer,
            layertype_to_replace,
            newlayer_constructor,
            suppress_inheritance=suppress_inheritance,
        )


def derive_args_from_layer(old_layer: nn.Module, new_layer_type, transformator=None):
    args: Dict[str, Any] = {}
    for p_name, param in signature(new_layer_type).parameters.items():
        # if param.default is param.empty:
        # First see if the old layer has the same attribute and take it
        # Then try to use the default
        p_old = getattr(old_layer, p_name, param.default)
        if True in [isinstance(p_old, t) for t in [str, int, tuple, bool]]:
            if transformator is not None:
                p_old = transformator(p_name, p_old)
            args[p_name] = p_old
        # except AttributeError as e:
        #     args.append(param.default)
    return args


def replace_tuples(layer_2d: nn.Module, target_layer_type) -> nn.Module:
    def two_tuple_to_three_tuple(p_name: str, p_old: Any) -> Any:
        if isinstance(p_old, tuple) and len(p_old) == 2:
            # print(f"REPLACING {p_name} with value {p_old}")
            return p_old[0], p_old[1], (p_old[0] + p_old[1]) // 2
        return p_old

    res = target_layer_type(**derive_args_from_layer(layer_2d, target_layer_type, two_tuple_to_three_tuple))

    # res.kernel_size = (
    #     conv2d_layer.kernel_size[0],
    #     conv2d_layer.kernel_size[1],
    #     conv2d_layer.kernel_size[0] + conv2d_layer.kernel_size[1] // 2
    # )
    # res.stride = (
    #     conv2d_layer.stride[0],
    #     conv2d_layer.stride[1],
    #     conv2d_layer.stride[0] + conv2d_layer.stride[1] // 2
    # )
    return res
    # return nn.Conv3d(
    #     conv2d_layer.in_channels,
    #     conv2d_layer.out_channels,
    #     # conv2d has a 2-tuple as kernel size which will result in an error
    #     kernel_size=conv2d_layer.kernel_size[0]
    # )


M = TypeVar("M", bound=nn.Module)


def convert_2d_to_3d(m: M) -> M:
    d: Dict[Type[nn.Module], Optional[Callable[[nn.Module], nn.Module]]] = {
        nn.Conv2d: lambda l: replace_tuples(l, nn.Conv3d),
        nn.AdaptiveAvgPool2d: lambda l: replace_tuples(l, nn.AdaptiveAvgPool3d),
        nn.Dropout2d: None,
        nn.InstanceNorm2d: None,
        nn.BatchNorm2d: lambda l: replace_tuples(l, nn.BatchNorm3d),
        nn.ConvTranspose2d: None,
        nn.MaxPool2d: lambda l: replace_tuples(l, nn.MaxPool3d),
        nn.AdaptiveMaxPool2d: None,
        nn.AvgPool2d: None,
        nn.ReplicationPad2d: None,
        nn.ConstantPad2d: None,
        # Batch norm does not need to be adapted when going to 3D?
        # StateLessBatchnorm: None
    }

    for old_layer, newlayer_constructor in d.items():
        if type(m) is old_layer:
            if newlayer_constructor is None:
                raise Exception(f"Layertype of {old_layer} is missing a way to convert to 3d!")
            return newlayer_constructor(m)
        replace_childen_recursive(m, old_layer, newlayer_constructor, suppress_inheritance=True)
    return m



"""
The line `import mmm.interactive as mtl` should be a time-efficient entry for interactive sessions.
"""

import mmm.interactive.blocks as blocks
import mmm.interactive.configs as configs
import mmm.interactive.data as data
import mmm.interactive.pipes as pipes
import mmm.interactive.tasks as tasks
import mmm.interactive.training as training
import mmm.interactive.utils as utils
